#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid at kovidgoyal.net>

import os
import re
import sys
import weakref
from collections import defaultdict
from collections.abc import Callable
from contextlib import contextmanager
from functools import partial
from threading import Thread
from typing import cast

from html5_parser import parse
from lxml import html
from qt.core import (
    QAction,
    QBrush,
    QByteArray,
    QCheckBox,
    QColor,
    QColorDialog,
    QDialog,
    QDialogButtonBox,
    QFont,
    QFontInfo,
    QFontMetrics,
    QFormLayout,
    QHBoxLayout,
    QIcon,
    QKeySequence,
    QLabel,
    QLineEdit,
    QMenu,
    QPalette,
    QPointF,
    QPushButton,
    QSize,
    QSpinBox,
    QSyntaxHighlighter,
    Qt,
    QTabWidget,
    QTextBlockFormat,
    QTextCharFormat,
    QTextCursor,
    QTextDocument,
    QTextEdit,
    QTextFormat,
    QTextFrameFormat,
    QTextImageFormat,
    QTextListFormat,
    QTimer,
    QToolButton,
    QUrl,
    QVBoxLayout,
    QWidget,
    pyqtSignal,
    pyqtSlot,
)

from calibre import browser, fit_image, xml_replace_entities
from calibre.db.constants import DATA_DIR_NAME
from calibre.ebooks.chardet import xml_to_unicode
from calibre.gui2 import (
    NO_URL_FORMATTING,
    FunctionDispatcher,
    choose_dir,
    choose_files,
    error_dialog,
    gprefs,
    is_dark_theme,
    local_path_for_resource,
    qapplication_or_fail,
    question_dialog,
    safe_open_url,
)
from calibre.gui2.book_details import resolved_css
from calibre.gui2.dialogs.progress import ProgressDialog
from calibre.gui2.flow_toolbar import create_flow_toolbar
from calibre.gui2.tweak_book.widgets import PlainTextEdit
from calibre.gui2.widgets import LineEditECM
from calibre.gui2.widgets2 import to_plain_text
from calibre.startup import connect_lambda
from calibre.utils.cleantext import clean_xml_chars
from calibre.utils.config import tweaks
from calibre.utils.filenames import make_long_path_useable
from calibre.utils.imghdr import what
from calibre.utils.localization import _

# Cleanup Qt markup {{{

OBJECT_REPLACEMENT_CHAR = '\ufffc'


def parse_style(style):
    props = filter(None, (x.strip() for x in style.split(';')))
    ans = {}
    for prop in props:
        try:
            k, v = prop.split(':', 1)
        except Exception:
            continue
        ans[k.strip().lower()] = v.strip()
    return ans


liftable_props = ('font-style', 'font-weight', 'font-family', 'font-size')


def lift_styles(tag, style_map):
    common_props = None
    has_text = bool(tag.text)
    child_styles = []
    for child in tag.iterchildren('*'):
        if child.tail and child.tail.strip():
            has_text = True
        style = style_map[child]
        child_styles.append(style)
        if not child.text and len(child) == 0:
            continue
        if common_props is None:
            common_props = style.copy()
        else:
            for k, v in tuple(common_props.items()):
                if style.get(k) != v:
                    del common_props[k]
    if not has_text and common_props:
        lifted_props = []
        tag_style = style_map[tag]
        for k in liftable_props:
            if k in common_props:
                lifted_props.append(k)
                tag_style[k] = common_props[k]
        if lifted_props:
            for style in child_styles:
                for k in lifted_props:
                    style.pop(k, None)


def filter_qt_styles(style):
    for k in tuple(style):
        # -qt-paragraph-type is a hack used by Qt for empty paragraphs
        if k.startswith('-qt-'):
            del style[k]


def remove_margins(tag, style):
    ml, mr, mt, mb = (style.pop('margin-' + k, None) for k in 'left right top bottom'.split())
    is_blockquote = ml == mr and ml and ml != '0px' and (ml != mt or ml != mb)
    if is_blockquote:
        tag.tag = 'blockquote'


def remove_zero_indents(style):
    ti = style.get('text-indent')
    if ti == '0px':
        del style['text-indent']


def remove_heading_font_styles(tag, style):
    lvl = int(tag.tag[1:])
    expected_size = (None, 'xx-large', 'x-large', 'large', None, 'small', 'x-small')[lvl]
    if style.get('font-size', 1) == expected_size:
        del style['font-size']
    if style.get('font-weight') in ('0', '600', '700', 'bold'):
        del style['font-weight']


def use_implicit_styling_for_span(span, style):
    is_italic = style.get('font-style') == 'italic'
    is_bold = style.get('font-weight') in ('600', '700', 'bold')
    if is_italic and not is_bold:
        del style['font-style']
        span.tag = 'em'
    elif is_bold and not is_italic:
        del style['font-weight']
        span.tag = 'strong'
    if span.tag == 'span' and style.get('text-decoration') == 'underline':
        span.tag = 'u'
        del style['text-decoration']
    if span.tag == 'span' and style.get('text-decoration') == 'line-through':
        span.tag = 's'
        del style['text-decoration']
    if span.tag == 'span' and style.get('vertical-align') in ('sub', 'super'):
        span.tag = 'sub' if style.pop('vertical-align') == 'sub' else 'sup'


def use_implicit_styling_for_a(a, style_map):
    for span in a.iterchildren('span'):
        style = style_map[span]
        if style.get('text-decoration') == 'underline':
            del style['text-decoration']
        if style.get('color') == '#0000ff':
            del style['color']
        break


def merge_contiguous_links(root):
    all_hrefs = set(root.xpath('//a/@href'))
    for href in all_hrefs:
        tags = root.xpath('//a[@href=$h]', h=href)
        processed = set()

        def insert_tag(parent, child):
            parent.tail = child.tail
            if child.text:
                children = parent.getchildren()
                if children:
                    children[-1].tail = (children[-1].tail or '') + child.text
                else:
                    parent.text = (parent.text or '') + child.text
            for gc in child.iterchildren('*'):
                parent.append(gc)

        for a in tags:
            if a in processed or a.tail:
                continue
            processed.add(a)
            n = a
            remove = []
            while not n.tail and n.getnext() is not None and getattr(n.getnext(), 'tag', None) == 'a' and n.getnext().get('href') == href:
                n = n.getnext()
                processed.add(n)
                remove.append(n)
            for n in remove:
                insert_tag(a, n)
                n.getparent().remove(n)


def convert_anchors_to_ids(root):
    anchors = root.xpath('//a[@name]')
    for a in anchors:
        p = a.getparent()
        if len(a.attrib) == 1 and not p.text and a is p[0] and not a.text and not p.get('id') and a.get('name') and len(a) == 0:
            p.text = a.tail
            p.set('id', a.get('name'))
            p.remove(a)


def cleanup_qt_markup(root):
    from calibre.ebooks.docx.cleanup import lift

    style_map = defaultdict(dict)
    for tag in root.xpath('//*[@style]'):
        style_map[tag] = parse_style(tag.get('style'))
    convert_anchors_to_ids(root)
    block_tags = root.xpath('//body/*')
    for tag in block_tags:
        lift_styles(tag, style_map)
        tag_style = style_map[tag]
        remove_margins(tag, tag_style)
        remove_zero_indents(tag_style)
        if tag.tag.startswith('h') and tag.tag[1:] in '123456':
            remove_heading_font_styles(tag, tag_style)
        for child in tag.iterdescendants('a'):
            use_implicit_styling_for_a(child, style_map)
        for child in tag.iterdescendants('span'):
            use_implicit_styling_for_span(child, style_map[child])
        if tag.tag == 'p' and style_map[tag].get('-qt-paragraph-type') == 'empty':
            del tag[:]
            tag.text = '\xa0'
        if tag.tag in ('ol', 'ul'):
            for li in tag.iterdescendants('li'):
                ts = style_map.get(li)
                if ts:
                    remove_margins(li, ts)
                    remove_zero_indents(ts)
    for img in root.xpath('//img[@style]'):
        s = style_map.get(img)
        if s:
            if s == {'float': 'left'}:
                s['margin-right'] = '0.5em'
            elif s == {'float': 'right'}:
                s['margin-left'] = '0.5em'
    for style in style_map.values():
        filter_qt_styles(style)
        fw = style.get('font-weight')
        if fw in ('600', '700'):
            style['font-weight'] = 'bold'
    for tag, style in style_map.items():
        if style:
            tag.set('style', '; '.join(f'{k}: {v}' for k, v in style.items()))
        else:
            tag.attrib.pop('style', None)
    for span in root.xpath('//span[not(@style)]'):
        lift(span)

    merge_contiguous_links(root)


# }}}


def fix_html(original_html, original_txt, remove_comments=True, callback=None):
    raw = original_html
    raw = xml_to_unicode(raw, strip_encoding_pats=True, resolve_entities=True)[0]
    if remove_comments:
        comments_pat = re.compile(r'<!--.*?-->', re.DOTALL)
        raw = comments_pat.sub('', raw)
    if not original_txt and '<img' not in raw.lower():
        return ''

    try:
        root = parse(raw, maybe_xhtml=False, sanitize_names=True)
    except Exception:
        root = parse(clean_xml_chars(raw), maybe_xhtml=False, sanitize_names=True)
    if root.xpath('//meta[@name="calibre-dont-sanitize"]'):
        # Bypass cleanup if special meta tag exists
        return original_html

    try:
        cleanup_qt_markup(root)
    except Exception:
        import traceback

        traceback.print_exc()
    if callback is not None:
        callback(root, original_txt)
    elems = []
    for body in root.xpath('//body'):
        if body.text:
            elems.append(body.text)
        elems += [html.tostring(x, encoding='unicode') for x in body if x.tag not in ('script', 'style')]

    if len(elems) > 1:
        ans = '<div>{}</div>'.format(''.join(elems))
    else:
        ans = ''.join(elems)
        if not ans.startswith('<'):
            ans = f'<p>{ans}</p>'
    return xml_replace_entities(ans)


class BlockAction(QAction):
    block_name: str = ''


class EditorWidget(QTextEdit, LineEditECM):  # {{{
    data_changed = pyqtSignal()
    insert_images_separately = False
    can_store_images = False

    action_bold: QAction
    action_italic: QAction
    action_underline: QAction
    action_strikethrough: QAction
    action_superscript: QAction
    action_subscript: QAction
    action_ordered_list: QAction
    action_unordered_list: QAction
    action_align_left: QAction
    action_align_center: QAction
    action_align_right: QAction
    action_align_justified: QAction
    action_undo: QAction
    action_redo: QAction
    action_remove_format: QAction
    action_copy: QAction
    action_paste: QAction
    action_paste_and_match_style: QAction
    action_cut: QAction
    action_indent: QAction
    action_outdent: QAction
    action_select_all: QAction
    action_color: QAction
    action_background: QAction
    action_insert_link: QAction
    action_insert_image: QAction
    action_insert_hr: QAction
    action_clear: QAction
    action_upper_case: QAction
    action_lower_case: QAction
    action_capitalize: QAction
    action_swap_case: QAction
    action_title_case: QAction
    action_block_style: QAction

    @property
    def readonly(self):
        return self.isReadOnly()

    @readonly.setter
    def readonly(self, val):
        self.setReadOnly(bool(val))

    @contextmanager
    def editing_cursor(self, set_cursor=True):
        c = self.textCursor()
        c.beginEditBlock()
        yield c
        c.endEditBlock()
        if set_cursor:
            self.setTextCursor(c)
        self.focus_self()

    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)
        self.setTabChangesFocus(True)
        doc = self.document()
        assert doc is not None
        doc.setDefaultStyleSheet(resolved_css() + '\n\nli { margin-top: 0.5ex; margin-bottom: 0.5ex; }')
        font = self.font()
        f = QFontInfo(font)
        delta = tweaks['change_book_details_font_size_by'] + 1
        if delta:
            font.setPixelSize(int(f.pixelSize() + delta))
            self.setFont(font)
        f = QFontMetrics(self.font())
        self.em_size = f.horizontalAdvance('m')
        self.base_url = None
        self._parent = weakref.ref(parent)
        self.shortcut_map = {}

        def r(name, icon, text, checkable=False, shortcut=None, callback=None):
            ac = QAction(QIcon.ic(icon + '.png'), text, self) if icon else QAction(text, self)
            ac.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
            if checkable:
                ac.setCheckable(checkable)
            setattr(self, 'action_' + name, ac)
            callback = callback or getattr(self, 'do_' + name)
            ac.triggered.connect(callback)
            if shortcut is not None:
                if isinstance(shortcut, str):
                    shortcut = QKeySequence(shortcut, QKeySequence.SequenceFormat.PortableText)
                self.shortcut_map[shortcut] = ac
                sc = shortcut if isinstance(shortcut, QKeySequence) else QKeySequence(shortcut)
                ac.setShortcut(sc)
                ac.setToolTip(text + f' [{sc.toString(QKeySequence.SequenceFormat.NativeText)}]')
            self.addAction(ac)

        r('bold', 'format-text-bold', _('Bold'), True, QKeySequence.StandardKey.Bold)
        r('italic', 'format-text-italic', _('Italic'), True, QKeySequence.StandardKey.Italic)
        r('underline', 'format-text-underline', _('Underline'), True, QKeySequence.StandardKey.Underline)
        r('strikethrough', 'format-text-strikethrough', _('Strikethrough'), True)
        r('superscript', 'format-text-superscript', _('Superscript'), True)
        r('subscript', 'format-text-subscript', _('Subscript'), True)
        r('ordered_list', 'format-list-ordered', _('Ordered list'), True)
        r('unordered_list', 'format-list-unordered', _('Unordered list'), True)

        r('align_left', 'format-justify-left', _('Align left'), True)
        r('align_center', 'format-justify-center', _('Align center'), True)
        r('align_right', 'format-justify-right', _('Align right'), True)
        r('align_justified', 'format-justify-fill', _('Align justified'), True)
        r('undo', 'edit-undo', _('Undo'), shortcut=QKeySequence.StandardKey.Undo)
        r('redo', 'edit-redo', _('Redo'), shortcut=QKeySequence.StandardKey.Redo)
        r('remove_format', 'edit-clear', _('Remove formatting'))
        r('copy', 'edit-copy', _('Copy'), shortcut=QKeySequence.StandardKey.Copy)
        r('paste', 'edit-paste', _('Paste'), shortcut=QKeySequence.StandardKey.Paste)
        r('paste_and_match_style', 'edit-paste', _('Paste and match style'), shortcut='ctrl+shift+v')
        r('cut', 'edit-cut', _('Cut'), shortcut=QKeySequence.StandardKey.Cut)
        r('indent', 'format-indent-more', _('Increase indentation'))
        r('outdent', 'format-indent-less', _('Decrease indentation'))
        r('select_all', 'edit-select-all', _('Select all'), shortcut=QKeySequence.StandardKey.SelectAll)

        r('color', 'format-text-color', _('Foreground color'))
        r('background', 'format-fill-color', _('Background color'))
        r(
            'insert_link',
            'insert-link',
            _('Insert link') if self.insert_images_separately else _('Insert link or image'),
            shortcut='Ctrl+l',
        )
        r('insert_image', 'view-image', _('Insert image'), shortcut='Ctrl+p')
        r(
            'insert_hr',
            'format-text-hr',
            _('Insert separator'),
        )
        r('clear', 'trash', _('Clear'))
        r('upper_case', '', _('Upper case'), shortcut='Ctrl+alt+u', callback=self.upper_case)
        r('lower_case', '', _('Lower case'), shortcut='Ctrl+alt+l', callback=self.lower_case)
        r('capitalize', '', _('Capitalize'), shortcut='Ctrl+alt+c', callback=self.capitalize)
        r('swap_case', '', _('Swap case'), shortcut='Ctrl+alt+s', callback=self.swap_case)
        r('title_case', '', _('Title case'), shortcut='Ctrl+alt+t', callback=self.title_case)

        self.action_block_style = QAction(QIcon.ic('format-text-heading.png'), _('Style text block'), self)
        self.action_block_style.setToolTip(_('Style the selected text block'))
        self.block_style_menu = QMenu(self)
        self.action_block_style.setMenu(self.block_style_menu)
        self.block_style_actions = []
        h = _('Heading {0}')
        for text, name in (
            (_('Normal'), 'p'),
            (h.format(1), 'h1'),
            (h.format(2), 'h2'),
            (h.format(3), 'h3'),
            (h.format(4), 'h4'),
            (h.format(5), 'h5'),
            (h.format(6), 'h6'),
            (_('Blockquote'), 'blockquote'),
        ):
            ac = BlockAction(text, self)
            self.block_style_menu.addAction(ac)
            ac.block_name = name
            ac.setCheckable(True)
            self.block_style_actions.append(ac)
            ac.triggered.connect(self.do_format_block)

        self.setHtml('')
        self.copyAvailable.connect(self.update_clipboard_actions)
        self.update_clipboard_actions(False)
        self.selectionChanged.connect(self.update_selection_based_actions)
        self.update_selection_based_actions()
        connect_lambda(self.undoAvailable, self, lambda self, yes: self.action_undo.setEnabled(yes))
        connect_lambda(self.redoAvailable, self, lambda self, yes: self.action_redo.setEnabled(yes))
        self.action_undo.setEnabled(False), self.action_redo.setEnabled(False)
        self.textChanged.connect(self.update_cursor_position_actions)
        self.cursorPositionChanged.connect(self.update_cursor_position_actions)
        self.textChanged.connect(self.data_changed)
        self.update_cursor_position_actions()

    def update_clipboard_actions(self, copy_available):
        self.action_copy.setEnabled(copy_available)
        self.action_cut.setEnabled(copy_available)

    def update_selection_based_actions(self):
        pass

    def update_cursor_position_actions(self):
        c = self.textCursor()
        tcf = c.charFormat()
        ls = c.currentList()
        self.action_ordered_list.setChecked(ls is not None and ls.format().style() == QTextListFormat.Style.ListDecimal)
        self.action_unordered_list.setChecked(ls is not None and ls.format().style() == QTextListFormat.Style.ListDisc)
        vert = tcf.verticalAlignment()
        self.action_superscript.setChecked(vert == QTextCharFormat.VerticalAlignment.AlignSuperScript)
        self.action_subscript.setChecked(vert == QTextCharFormat.VerticalAlignment.AlignSubScript)
        self.action_bold.setChecked(tcf.fontWeight() == QFont.Weight.Bold)
        self.action_italic.setChecked(tcf.fontItalic())
        self.action_underline.setChecked(tcf.fontUnderline())
        self.action_strikethrough.setChecked(tcf.fontStrikeOut())
        bf = c.blockFormat()
        a = bf.alignment()
        self.action_align_left.setChecked(a == Qt.AlignmentFlag.AlignLeft)
        self.action_align_right.setChecked(a == Qt.AlignmentFlag.AlignRight)
        self.action_align_center.setChecked(a == Qt.AlignmentFlag.AlignHCenter)
        self.action_align_justified.setChecked(a == Qt.AlignmentFlag.AlignJustify)
        lvl = bf.headingLevel()
        name = 'p'
        if lvl == 0:
            if bf.leftMargin() == bf.rightMargin() and bf.leftMargin() > 0:
                name = 'blockquote'
        else:
            name = f'h{lvl}'
        for ac in self.block_style_actions:
            ac.setChecked(ac.block_name == name)

    def set_readonly(self, what):
        self.readonly = what

    def focus_self(self):
        self.setFocus(Qt.FocusReason.TabFocusReason)

    def do_clear(self, *args):
        c = self.textCursor()
        c.beginEditBlock()
        c.movePosition(QTextCursor.MoveOperation.Start, QTextCursor.MoveMode.MoveAnchor)
        c.movePosition(QTextCursor.MoveOperation.End, QTextCursor.MoveMode.KeepAnchor)
        c.removeSelectedText()
        c.endEditBlock()
        self.focus_self()

    clear_text = do_clear

    def do_bold(self):
        with self.editing_cursor() as c:
            fmt = QTextCharFormat()
            fmt.setFontWeight(QFont.Weight.Bold if c.charFormat().fontWeight() != QFont.Weight.Bold else QFont.Weight.Normal)
            c.mergeCharFormat(fmt)
        self.update_cursor_position_actions()

    def do_italic(self):
        with self.editing_cursor() as c:
            fmt = QTextCharFormat()
            fmt.setFontItalic(not c.charFormat().fontItalic())
            c.mergeCharFormat(fmt)
        self.update_cursor_position_actions()

    def do_underline(self):
        with self.editing_cursor() as c:
            fmt = QTextCharFormat()
            fmt.setFontUnderline(not c.charFormat().fontUnderline())
            c.mergeCharFormat(fmt)
        self.update_cursor_position_actions()

    def do_strikethrough(self):
        with self.editing_cursor() as c:
            fmt = QTextCharFormat()
            fmt.setFontStrikeOut(not c.charFormat().fontStrikeOut())
            c.mergeCharFormat(fmt)
        self.update_cursor_position_actions()

    def do_vertical_align(self, which):
        with self.editing_cursor() as c:
            fmt = QTextCharFormat()
            fmt.setVerticalAlignment(which)
            c.mergeCharFormat(fmt)
        self.update_cursor_position_actions()

    def do_superscript(self):
        self.do_vertical_align(QTextCharFormat.VerticalAlignment.AlignSuperScript)

    def do_subscript(self):
        self.do_vertical_align(QTextCharFormat.VerticalAlignment.AlignSubScript)

    def do_list(self, fmt):
        with self.editing_cursor() as c:
            ls = c.currentList()
            if ls is not None:
                lf = ls.format()
                if lf.style() == fmt:
                    c.setBlockFormat(QTextBlockFormat())
                else:
                    lf.setStyle(fmt)
                    ls.setFormat(lf)
            else:
                ls = c.createList(fmt)
        self.update_cursor_position_actions()

    def do_ordered_list(self):
        self.do_list(QTextListFormat.Style.ListDecimal)

    def do_unordered_list(self):
        self.do_list(QTextListFormat.Style.ListDisc)

    def do_alignment(self, which):
        with self.editing_cursor() as c:
            c = self.textCursor()
            fmt = QTextBlockFormat()
            fmt.setAlignment(which)
            c.mergeBlockFormat(fmt)
        self.update_cursor_position_actions()

    def do_align_left(self):
        self.do_alignment(Qt.AlignmentFlag.AlignLeft)

    def do_align_center(self):
        self.do_alignment(Qt.AlignmentFlag.AlignHCenter)

    def do_align_right(self):
        self.do_alignment(Qt.AlignmentFlag.AlignRight)

    def do_align_justified(self):
        self.do_alignment(Qt.AlignmentFlag.AlignJustify)

    def do_undo(self):
        self.undo()
        self.focus_self()

    def do_redo(self):
        self.redo()
        self.focus_self()

    def do_remove_format(self):
        with self.editing_cursor() as c:
            c.setBlockFormat(QTextBlockFormat())
            c.setCharFormat(QTextCharFormat())
        self.update_cursor_position_actions()

    def keyPressEvent(self, e):
        for sc, ac in self.shortcut_map.items():
            if isinstance(sc, QKeySequence.StandardKey) and e.matches(sc):
                ac.trigger()
                return
        return super().keyPressEvent(e)

    def do_copy(self):
        self.copy()
        self.focus_self()

    def do_paste(self):
        images_before = set()
        doc = self.document()
        assert doc is not None
        for fmt in doc.allFormats():
            if fmt.isImageFormat():
                images_before.add(fmt.toImageFormat().name())
        self.paste()
        if self.can_store_images:
            added = set()
            for fmt in doc.allFormats():
                if fmt.isImageFormat():
                    name = fmt.toImageFormat().name()
                    if name not in images_before and name.partition(':')[0] in ('https', 'http'):
                        added.add(name)
            if added and question_dialog(self, _('Download images'), _('Download all remote images in the pasted content?')):
                self.download_images(added)
        self.focus_self()

    def commit_downloaded_image(self, data: bytes, suggested_filename: str) -> str:
        raise NotImplementedError

    def download_images(self, urls):
        from calibre.web import get_download_filename_from_response

        br = browser()
        d = self.document()
        assert d is not None
        c = self.textCursor()
        c.setPosition(0)
        pos_map = {}
        while True:
            c = d.find(OBJECT_REPLACEMENT_CHAR, c, QTextDocument.FindFlag.FindCaseSensitively)
            if c.isNull():
                break
            fmt = c.charFormat()
            if fmt.isImageFormat():
                name = fmt.toImageFormat().name()
                if name in urls:
                    pos_map[name] = c.position()

        pd = ProgressDialog(title=_('Downloading images...'), min=0, max=0, parent=self)
        pd.canceled_signal.connect(pd.accept)
        data_map = {}
        error_map = {}
        set_msg = FunctionDispatcher(pd.set_msg, parent=pd)
        accept = FunctionDispatcher(pd.accept, parent=pd)

        def do_download():
            for url, pos in pos_map.items():
                if pd.canceled:
                    return
                set_msg(_('Downloading {}...').format(url))
                try:
                    res = br.open_novisit(url, timeout=30)
                    fname = get_download_filename_from_response(res)
                    data = res.read()
                except Exception as err:
                    error_map[url] = err
                else:
                    data_map[url] = fname, data
            accept()

        Thread(target=do_download, daemon=True).start()
        pd.exec()
        if pd.canceled:
            return

        for url, pos in pos_map.items():
            fname, data = data_map[url]
            newname = self.commit_downloaded_image(data, fname)
            c = self.textCursor()
            c.setPosition(pos)
            alignment = QTextFrameFormat.Position.InFlow
            f = self.frame_for_cursor(c)
            if f is not None:
                alignment = f.frameFormat().position()
            fmt = c.charFormat()
            i = fmt.toImageFormat()
            i.setName(newname)
            c.deletePreviousChar()
            c.insertImage(i, alignment)
        if error_map:
            m = '\n'.join(_('Could not download {0} with error:').format(url) + '\n\t' + str(err) + '\n\n' for url, err in error_map.items())
            error_dialog(
                self,
                _('Failed to download some images'),
                _('Some images could not be downloaded, click "Show details" to see which ones'),
                det_msg=m,
                show=True,
            )

    def do_paste_and_match_style(self):
        clipboard = qapplication_or_fail().clipboard()
        assert clipboard is not None
        text = clipboard.text()
        if text:
            self.setText(text)

    def do_cut(self):
        self.cut()
        self.focus_self()

    def indent_block(self, mult=1):
        with self.editing_cursor() as c:
            bf = c.blockFormat()
            bf.setTextIndent(bf.textIndent() + 2 * self.em_size * mult)
            c.setBlockFormat(bf)
        self.update_cursor_position_actions()

    def do_indent(self):
        self.indent_block()

    def do_outdent(self):
        self.indent_block(-1)

    def do_select_all(self):
        with self.editing_cursor() as c:
            c.movePosition(QTextCursor.MoveOperation.Start, QTextCursor.MoveMode.MoveAnchor)
            c.movePosition(QTextCursor.MoveOperation.End, QTextCursor.MoveMode.KeepAnchor)

    def level_for_block_type(self, name):
        if name == 'blockquote':
            return 0
        return {q: i for i, q in enumerate('p h1 h2 h3 h4 h5 h6'.split())}[name]

    def do_format_block(self):
        sender = self.sender()
        assert isinstance(sender, BlockAction)
        name = sender.block_name
        with self.editing_cursor() as c:
            bf = QTextBlockFormat()
            cf = QTextCharFormat()
            bcf = c.blockCharFormat()
            lvl = self.level_for_block_type(name)
            wt = QFont.Weight.Bold if lvl else QFont.Weight.Normal
            adjust = (0, 3, 2, 1, 0, -1, -1)[lvl]
            pos = None
            if not c.hasSelection():
                pos = c.position()
                c.movePosition(QTextCursor.MoveOperation.StartOfBlock, QTextCursor.MoveMode.MoveAnchor)
                c.movePosition(QTextCursor.MoveOperation.EndOfBlock, QTextCursor.MoveMode.KeepAnchor)
            # margin values are taken from qtexthtmlparser.cpp
            hmargin = 0
            if name == 'blockquote':
                hmargin = 40
            tmargin = bmargin = 12
            if name == 'h1':
                tmargin, bmargin = 18, 12
            elif name == 'h2':
                tmargin, bmargin = 16, 12
            elif name == 'h3':
                tmargin, bmargin = 14, 12
            elif name == 'h4':
                tmargin, bmargin = 12, 12
            elif name == 'h5':
                tmargin, bmargin = 12, 4
            bf.setLeftMargin(hmargin), bf.setRightMargin(hmargin)
            bf.setTopMargin(tmargin), bf.setBottomMargin(bmargin)
            bf.setHeadingLevel(lvl)
            bcf.setProperty(QTextFormat.Property.FontSizeAdjustment, adjust)
            cf.setProperty(QTextFormat.Property.FontSizeAdjustment, adjust)
            bcf.setFontWeight(wt)
            cf.setFontWeight(wt)
            c.setBlockCharFormat(bcf)
            c.mergeCharFormat(cf)
            c.mergeBlockFormat(bf)
            if pos is not None:
                c.setPosition(pos)
        self.update_cursor_position_actions()

    def do_color(self):
        col = QColorDialog.getColor(Qt.GlobalColor.black, self, _('Choose foreground color'), QColorDialog.ColorDialogOption.ShowAlphaChannel)
        if col.isValid():
            fmt = QTextCharFormat()
            fmt.setForeground(QBrush(col))
            with self.editing_cursor() as c:
                c.mergeCharFormat(fmt)

    def do_background(self):
        col = QColorDialog.getColor(Qt.GlobalColor.white, self, _('Choose background color'), QColorDialog.ColorDialogOption.ShowAlphaChannel)
        if col.isValid():
            fmt = QTextCharFormat()
            fmt.setBackground(QBrush(col))
            with self.editing_cursor() as c:
                c.mergeCharFormat(fmt)

    def do_insert_hr(self, *args):
        with self.editing_cursor() as c:
            c.movePosition(QTextCursor.MoveOperation.EndOfBlock, QTextCursor.MoveMode.MoveAnchor)
            c.insertHtml('<hr>')

    def do_insert_image(self):
        from calibre.gui2 import choose_images

        files = choose_images(self, 'choose-image-for-comments-editor', _('Choose image'), formats='png jpeg jpg gif svg webp'.split())
        if files:
            self.focus_self()
            with self.editing_cursor() as c:
                c.insertImage(files[0])

    def do_insert_link(self, *args):
        link, name, is_image = self.ask_link()
        if not link:
            return
        url = self.parse_link(link)
        if url.isValid():
            if url.isLocalFile() and not os.path.isabs(url.toLocalFile()):
                url = url.toLocalFile()
            else:
                url = url.toString(NO_URL_FORMATTING)
            self.focus_self()
            with self.editing_cursor() as c:
                if is_image:
                    c.insertImage(url)
                else:
                    oldfmt = QTextCharFormat(c.charFormat())
                    fmt = QTextCharFormat()
                    fmt.setAnchor(True)
                    fmt.setAnchorHref(url)
                    fmt.setForeground(QBrush(self.palette().color(QPalette.ColorRole.Link)))
                    if name or not c.hasSelection():
                        c.mergeCharFormat(fmt)
                        c.insertText(name or url)
                    else:
                        pos, anchor = c.position(), c.anchor()
                        start, end = min(pos, anchor), max(pos, anchor)
                        for i in range(start, end):
                            cur = self.textCursor()
                            cur.setPosition(i), cur.setPosition(i + 1, QTextCursor.MoveMode.KeepAnchor)
                            cur.mergeCharFormat(fmt)
                    c.setPosition(c.position())
                    c.setCharFormat(oldfmt)

        else:
            error_dialog(self, _('Invalid URL'), _('The URL %r is invalid') % link, show=True)

    def ask_link(self):

        class Ask(QDialog):
            url: QLineEdit
            name: QLineEdit
            treat_as_image: QCheckBox
            bb: QDialogButtonBox
            br: QPushButton
            brdf: QPushButton
            brd: QPushButton
            la: QLabel

            def accept(self):
                if self.treat_as_image.isChecked():
                    url = self.url.text()
                    if url.lower().split(':', 1)[0] in ('http', 'https'):
                        error_dialog(
                            self,
                            _('Remote images not supported'),
                            _('You must download the image to your computer, URLs pointing to remote images are not supported.'),
                            show=True,
                        )
                        return
                QDialog.accept(self)

        d = Ask(self)
        d.setWindowTitle(_('Create link'))
        l = QFormLayout()
        l.setFieldGrowthPolicy(QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        d.setLayout(l)
        d.url = QLineEdit(d)
        d.name = QLineEdit(d)
        d.treat_as_image = QCheckBox(d)
        d.setMinimumWidth(600)
        d.bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        d.br = b = QPushButton(_('&File'))
        base = os.path.dirname(self.base_url.toLocalFile()) if self.base_url else os.getcwd()
        data_path = os.path.join(base, DATA_DIR_NAME)
        if self.base_url:
            os.makedirs(data_path, exist_ok=True)

        def cf(data_dir=False):
            filetypes = []
            if d.treat_as_image.isChecked():
                filetypes = [(_('Images'), 'png jpeg jpg gif'.split())]
            if data_dir:
                files = choose_files(
                    d,
                    'select link file',
                    _('Choose file'),
                    filetypes,
                    select_only_single_file=True,
                    no_save_dir=True,
                    default_dir=data_path,
                )
            else:
                files = choose_files(d, 'select link file', _('Choose file'), filetypes, select_only_single_file=True)
            if files:
                path = files[0]
                d.url.setText(path)
                if path and os.path.exists(path):
                    with open(path, 'rb') as f:
                        q = what(f)
                    is_image = q in {'jpeg', 'png', 'gif', 'webp'}
                    d.treat_as_image.setChecked(is_image)
                if data_dir:
                    path = os.path.relpath(path, base).replace(os.sep, '/')
                    d.url.setText(path)

        b.clicked.connect(cf)
        d.brdf = b = QPushButton(_('&Data file'))
        b.clicked.connect(lambda: cf(True))
        b.setToolTip(_('A relative link to a data file associated with this book'))
        if not os.path.exists(data_path):
            b.setVisible(False)
        d.brd = b = QPushButton(_('F&older'))

        def cd():
            path = choose_dir(d, 'select link folder', _('Choose folder'))
            if path:
                d.url.setText(path)

        b.clicked.connect(cd)

        d.la = la = QLabel(
            _(
                'Enter a URL. If you check the "Treat the URL as an image" box '
                'then the URL will be added as an image reference instead of as '
                'a link. You can also choose to create a link to a file on '
                'your computer. '
                'Note that if you create a link to a file on your computer, it '
                'will stop working if the file is moved.'
            )
        )
        la.setWordWrap(True)
        la.setStyleSheet('QLabel { margin-bottom: 1.5ex }')
        l.setWidget(0, QFormLayout.ItemRole.SpanningRole, la)
        l.addRow(_('Enter &URL:'), d.url)
        l.addRow(_('Treat the URL as an &image'), d.treat_as_image)
        l.addRow(_('Enter &name (optional):'), d.name)
        h = QHBoxLayout()
        h.addWidget(d.br), h.addWidget(d.brdf), h.addWidget(d.brd)
        l.addRow(_('Choose a file on your computer:'), h)
        l.addRow(d.bb)
        d.bb.accepted.connect(d.accept)
        d.bb.rejected.connect(d.reject)
        d.resize(d.sizeHint())
        link, name, is_image = None, None, False
        if d.exec() == QDialog.DialogCode.Accepted:
            link, name = str(d.url.text()).strip(), str(d.name.text()).strip()
            is_image = d.treat_as_image.isChecked()
        return link, name, is_image

    def parse_link(self, link):
        link = link.strip()
        if link and os.path.exists(link):
            return QUrl.fromLocalFile(link)
        has_schema = re.match(r'^[a-zA-Z]+:', link)
        if has_schema is not None:
            url = QUrl(link, QUrl.ParsingMode.TolerantMode)
            if url.isValid():
                return url
        else:
            if self.base_url:
                base = os.path.dirname(self.base_url.toLocalFile())
            else:
                base = os.getcwd()
            candidate = os.path.join(base, link)
            if os.path.exists(candidate):
                return QUrl.fromLocalFile(link)
        if os.path.exists(link):
            return QUrl.fromLocalFile(link)

        if has_schema is None:
            first, _, rest = link.partition('.')
            prefix = 'http'
            if first == 'ftp':
                prefix = 'ftp'
            url = QUrl(prefix + '://' + link, QUrl.ParsingMode.TolerantMode)
            if url.isValid():
                return url

        return QUrl(link, QUrl.ParsingMode.TolerantMode)

    def sizeHint(self):
        return QSize(150, 150)

    @property
    def html(self):
        return fix_html(self.toHtml(), self.toPlainText().strip(), callback=self.get_html_callback)

    def get_html_callback(self, root, text):
        pass

    @html.setter
    def html(self, val):
        self.setHtml(val)

    def set_base_url(self, qurl):
        self.base_url = qurl

    @pyqtSlot(int, 'QUrl', result='QVariant')
    def loadResource(self, type, name):
        path = local_path_for_resource(name, base_qurl=self.base_url)
        if path:
            data = None
            try:
                with open(make_long_path_useable(path), 'rb') as f:
                    data = f.read()
            except OSError:
                if path.rpartition('.')[-1].lower() in {'jpg', 'jpeg', 'gif', 'png', 'bmp', 'webp'}:
                    data = bytearray.fromhex(
                        '89504e470d0a1a0a0000000d49484452000000010000000108060000001f15c4890000000a49444154789c63000100000500010d0a2db40000000049454e44ae426082'
                    )
            if data is not None:
                r = QByteArray(data)
                doc = self.document()
                assert doc is not None
                doc.addResource(type, name, r)
                return r

    def set_html(self, val, allow_undo=True):
        if not allow_undo or self.readonly:
            self.html = val
            return
        with self.editing_cursor() as c:
            c.movePosition(QTextCursor.MoveOperation.Start, QTextCursor.MoveMode.MoveAnchor)
            c.movePosition(QTextCursor.MoveOperation.End, QTextCursor.MoveMode.KeepAnchor)
            c.removeSelectedText()
            c.insertHtml(val)

    def text(self) -> str:
        return self.textCursor().selectedText()

    def selectedText(self) -> str:
        return self.text()

    def setText(self, text: str | None) -> None:
        with self.editing_cursor() as c:
            c.insertText(text or '')

    def insert(self, a0: str | None) -> None:
        self.setText(a0)

    def hasSelectedText(self) -> bool:
        c = self.textCursor()
        return c.hasSelection()

    def createMimeDataFromSelection(self):
        ans = super().createMimeDataFromSelection()
        assert ans is not None
        html, txt = ans.html(), ans.text()
        html = fix_html(html, txt, remove_comments=False)
        # Qt has a bug where copying from the start of a paragraph does not
        # include the paragraph definition in the fragment. Try to fix that
        # by moving the StartFragment comment to before the paragraph when
        # the selection starts at the start of a block
        c = self.textCursor()
        c2 = QTextCursor(c)
        c2.setPosition(c.selectionStart())
        if c2.atBlockStart():
            html = re.sub(r'(<p.*?>)(<!--StartFragment-->)', r'\2\1', html, count=1)
        ans.setHtml(html)
        return ans

    def remove_image_at(self, cursor_pos):
        c = self.textCursor()
        c.clearSelection()
        c.setPosition(cursor_pos)
        fmt = c.charFormat()
        if fmt.isImageFormat():
            c.deletePreviousChar()

    def resize_image_at(self, cursor_pos):
        c = self.textCursor()
        c.clearSelection()
        c.setPosition(cursor_pos)
        c.movePosition(QTextCursor.MoveOperation.PreviousCharacter, QTextCursor.MoveMode.KeepAnchor)
        fmt = c.charFormat()
        if not fmt.isImageFormat():
            return
        fmt = fmt.toImageFormat()
        from calibre.utils.img import image_from_data

        img = image_from_data(self.loadResource(QTextDocument.ResourceType.ImageResource, QUrl(fmt.name())))
        w, h = int(fmt.width()), int(fmt.height())
        d = QDialog(self)
        l = QVBoxLayout(d)
        la = QLabel(_('Shrink image to fit within:'))
        h = QHBoxLayout()
        l.addLayout(h)
        la = QLabel(_('&Width:'))
        h.addWidget(la)
        width_spinbox = w = QSpinBox(self)
        w.setRange(0, 10000), w.setSuffix(' px')
        w.setValue(int(fmt.width()))
        h.addWidget(w), la.setBuddy(w)
        w.setSpecialValueText(' ')
        la = QLabel(_('&Height:'))
        h.addWidget(la)
        height_spinbox = w = QSpinBox(self)
        w.setRange(0, 10000), w.setSuffix(' px')
        w.setValue(int(fmt.height()))
        h.addWidget(w), la.setBuddy(w)
        w.setSpecialValueText(' ')
        h.addStretch(10)
        bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        bb.accepted.connect(d.accept), bb.rejected.connect(d.reject)
        d.setWindowTitle(_('Enter new size for image'))
        l.addWidget(bb)
        d.resize(d.sizeHint())

        if d.exec() == QDialog.DialogCode.Accepted:
            page_width, page_height = (width_spinbox.value() or sys.maxsize), (height_spinbox.value() or sys.maxsize)
            w, h = int(img.width()), int(img.height())
            resized, nw, nh = fit_image(w, h, page_width, page_height)
            if resized:
                fmt.setWidth(nw), fmt.setHeight(nh)
            else:
                f = QTextImageFormat()
                f.setName(fmt.name())
                fmt = f
            c.setCharFormat(fmt)

    def align_image_at(self, cursor_pos, alignment):
        c = self.textCursor()
        c.clearSelection()
        c.setPosition(cursor_pos)
        fmt = c.charFormat()
        if fmt.isImageFormat() and c.currentFrame():
            f = self.frame_for_cursor(c)
            if f is not None:
                ff = f.frameFormat()
                ff.setPosition(alignment)
                f.setFrameFormat(ff)
                doc = self.document()
                assert doc is not None
                doc.markContentsDirty(cursor_pos - 2, 5)
            else:
                c.deleteChar()
                c.insertImage(fmt.toImageFormat(), alignment)

    def first_image_replacement_char_position_for(self, image_name):
        d = self.document()
        assert d is not None
        c = self.textCursor()
        c.setPosition(0)
        while True:
            c = d.find(OBJECT_REPLACEMENT_CHAR, c, QTextDocument.FindFlag.FindCaseSensitively)
            if c.isNull():
                break
            fmt = c.charFormat()
            if fmt.isImageFormat() and fmt.toImageFormat().name() == image_name:
                return c.position()
        return -1

    def frame_for_cursor(self, c):
        q = c.position()
        for cf in c.currentFrame().childFrames():
            a, b = cf.firstPosition(), cf.lastPosition()
            a, b = min(a, b), max(a, b)
            if a <= q <= b:
                return cf

    def open_link(self, link: str) -> None:
        name = QUrl(link, QUrl.ParsingMode.TolerantMode)
        if name.isRelative() and self.base_url:
            name = QUrl.fromLocalFile(os.path.join(os.path.dirname(self.base_url.toLocalFile()), name.path()))
        safe_open_url(name)

    def contextMenuEvent(self, e):
        menu = QMenu(self)
        doc = self.document()
        assert doc is not None
        doc_layout = doc.documentLayout()
        assert doc_layout is not None
        img_name = doc_layout.imageAt(QPointF(e.pos()))
        if img_name:
            pos = self.first_image_replacement_char_position_for(img_name)
            if pos > -1:
                c = self.textCursor()
                c.clearSelection()
                c.setPosition(pos)
                pos = QTextFrameFormat.Position.InFlow
                ff = self.frame_for_cursor(c)
                if ff is not None:
                    pos = ff.frameFormat().position()
                align_menu = menu.addMenu(QIcon.ic('view-image.png'), _('Image...'))
                assert align_menu is not None

                def a(text, epos):
                    ac = align_menu.addAction(text)
                    ac.setCheckable(True)
                    ac.triggered.connect(partial(self.align_image_at, c.position(), epos))
                    if pos == epos:
                        ac.setChecked(True)

                cs = align_menu.addAction(QIcon.ic('resize.png'), _('Change size'))
                assert cs is not None
                cs.triggered.connect(partial(self.resize_image_at, c.position()))
                align_menu.addSeparator()
                a(_('Float to the left'), QTextFrameFormat.Position.FloatLeft)
                a(_('Inline with text'), QTextFrameFormat.Position.InFlow)
                a(_('Float to the right'), QTextFrameFormat.Position.FloatRight)
                align_menu.addSeparator()
                trash_action = align_menu.addAction(QIcon.ic('trash.png'), _('Remove this image'))
                assert trash_action is not None
                trash_action.triggered.connect(partial(self.remove_image_at, c.position()))
        link_name = doc_layout.anchorAt(QPointF(e.pos()))
        if link_name:
            menu.addAction(QIcon.ic('insert-link.png'), _('Open link'), partial(self.open_link, link_name))
        for ac in 'undo redo -- cut copy paste paste_and_match_style -- select_all'.split():
            if ac == '--':
                menu.addSeparator()
            else:
                ac = getattr(self, 'action_' + ac)
                menu.addAction(ac)
        st = self.text()
        m = QMenu(_('Fonts'))
        m.addAction(self.action_bold), m.addAction(self.action_italic), m.addAction(self.action_underline)
        menu.addMenu(m)

        if st and st.strip():
            m = QMenu(_('Change case'), menu)
            m.addAction(self.action_upper_case)
            m.addAction(self.action_lower_case)
            m.addAction(self.action_swap_case)
            m.addAction(self.action_title_case)
            m.addAction(self.action_capitalize)
            menu.addMenu(m)
        parent = self._parent()
        assert parent is not None
        if hasattr(parent, 'toolbars_visible'):
            vis = parent.toolbars_visible
            menu.addAction(_('%s toolbars') % (_('Hide') if vis else _('Show')), parent.toggle_toolbars)
        menu.addSeparator()
        am = QMenu(_('Advanced'))
        menu.addMenu(am)
        am.addAction(self.action_block_style)
        am.addAction(self.action_insert_link)
        if self.insert_images_separately:
            am.addAction(self.action_insert_image)
        am.addAction(self.action_background)
        am.addAction(self.action_color)
        menu.addAction(_('Smarten punctuation'), parent.smarten_punctuation)
        menu.exec(e.globalPos())

    def modify_case_operation(self, func: Callable[[str], str]) -> None:
        cursor: QTextCursor = self.textCursor()
        if not cursor.hasSelection():
            cursor.select(QTextCursor.SelectionType.Document)
        start = cursor.selectionStart()
        end = cursor.selectionEnd()
        doc = self.document()

        # Helper: collect blocks intersecting the selection
        blocks = []

        # Create a cursor at start to find the first block
        sc = QTextCursor(doc)
        sc.setPosition(start)
        first_block = sc.block()

        block = first_block
        while block.isValid() and block.position() < end:
            block_start = block.position()
            block_len = block.length()  # includes block separator
            block_end = block_start + block_len
            seg_start = max(start, block_start)
            seg_end = min(end, block_end)
            if seg_start >= seg_end:
                block = block.next()
                continue

            # record block-level formatting (block format and list format if any)
            bf = QTextBlockFormat(block.blockFormat())
            list_format = None
            blist = block.textList()
            if blist is not None:
                list_format = QTextListFormat(blist.format())

            # iterate fragments in the block and extract only the overlapped text pieces
            fragments = []
            fragment_text = ''
            it = block.begin()
            while not it.atEnd():
                frag = it.fragment()
                if frag.isValid():
                    fpos = frag.position()
                    flen = frag.length()
                    fend = fpos + flen
                    ov_start = max(seg_start, fpos)
                    ov_end = min(seg_end, fend)
                    if ov_start < ov_end:
                        rel_start = ov_start - fpos
                        rel_len = ov_end - ov_start
                        text = frag.text()[rel_start : rel_start + rel_len]
                        fmt = QTextCharFormat(frag.charFormat())
                        fragments.append((len(fragment_text), len(text), fmt))
                        fragment_text += text
                it += 1

            blocks.append({
                'blockFormat': bf,
                'listFormat': list_format,
                'blockStart': block_start,
                'fragments': fragments,
                'fragment_text': fragment_text,
            })

            block = block.next()

        if not blocks:
            return

        # Replace the selection preserving inline and block formats.
        editcur = QTextCursor(doc)
        editcur.setPosition(start)
        editcur.setPosition(end, QTextCursor.MoveMode.KeepAnchor)

        editcur.beginEditBlock()
        editcur.removeSelectedText()
        # after removal the cursor is at `start`
        editcur.setPosition(start)

        # Insert blocks and fragments. Handle first block specially to avoid unwanted new block
        for idx, blk in enumerate(blocks):
            blk_fmt = cast(QTextBlockFormat, blk['blockFormat'])
            list_fmt = blk['listFormat']

            # Determine if original first block was partial (selection started mid-block).
            if idx == 0:
                # keep and reuse the current block — apply its block format/list
                editcur.setBlockFormat(blk_fmt)
            else:
                # subsequent blocks -> always insert a new block with the original format
                editcur.insertBlock(blk_fmt)
            if isinstance(list_fmt, QTextListFormat):
                editcur.createList(list_fmt)

            fragment_text = func(cast(str, blk['fragment_text']))
            # insert the fragments for this block, preserving char formats
            for start_pos, length, ch_fmt in cast(list[tuple[int, int, QTextCharFormat]], blk['fragments']):
                if start_pos >= len(fragment_text):
                    break
                # Convert selected fragment text
                # (paragraph separators are not part of per-fragment text here)
                editcur.insertText(fragment_text[start_pos : start_pos + length], ch_fmt)

        editcur.endEditBlock()

        # Reselect the replaced text so selection remains
        new_end = editcur.position()
        new_cursor = self.textCursor()
        new_cursor.setPosition(start)
        new_cursor.setPosition(new_end, QTextCursor.MoveMode.KeepAnchor)
        self.setTextCursor(new_cursor)


# }}}

# Highlighter {{{

State_Text = -1
State_DocType = 0
State_Comment = 1
State_TagStart = 2
State_TagName = 3
State_InsideTag = 4
State_AttributeName = 5
State_SingleQuote = 6
State_DoubleQuote = 7
State_AttributeValue = 8


class Highlighter(QSyntaxHighlighter):
    def __init__(self, doc):
        QSyntaxHighlighter.__init__(self, doc)
        self.colors = {}
        self.colors['doctype'] = QColor(192, 192, 192)
        self.colors['entity'] = QColor(128, 128, 128)
        self.colors['comment'] = QColor(35, 110, 37)
        if is_dark_theme():
            from calibre.gui2.palette import dark_link_color

            self.colors['tag'] = QColor(186, 78, 188)
            self.colors['attrname'] = QColor(193, 119, 60)
            self.colors['attrval'] = dark_link_color
        else:
            self.colors['tag'] = QColor(136, 18, 128)
            self.colors['attrname'] = QColor(153, 69, 0)
            self.colors['attrval'] = QColor(36, 36, 170)

    def highlightBlock(self, text):
        state = self.previousBlockState()
        len_ = len(text)
        start = 0
        pos = 0

        while pos < len_:
            if state == State_Comment:
                start = pos
                while pos < len_:
                    if text[pos : pos + 3] == '-->':
                        pos += 3
                        state = State_Text
                        break
                    else:
                        pos += 1
                self.setFormat(start, pos - start, self.colors['comment'])

            elif state == State_DocType:
                start = pos
                while pos < len_:
                    ch = text[pos]
                    pos += 1
                    if ch == '>':
                        state = State_Text
                        break
                self.setFormat(start, pos - start, self.colors['doctype'])

            # at '<' in e.g. "<span>foo</span>"
            elif state == State_TagStart:
                start = pos + 1
                while pos < len_:
                    ch = text[pos]
                    pos += 1
                    if ch == '>':
                        state = State_Text
                        break
                    if not ch.isspace():
                        pos -= 1
                        state = State_TagName
                        break

            # at 'b' in e.g "<blockquote>foo</blockquote>"
            elif state == State_TagName:
                start = pos
                while pos < len_:
                    ch = text[pos]
                    pos += 1
                    if ch.isspace():
                        pos -= 1
                        state = State_InsideTag
                        break
                    if ch == '>':
                        state = State_Text
                        break
                self.setFormat(start, pos - start, self.colors['tag'])

            # anywhere after tag name and before tag closing ('>')
            elif state == State_InsideTag:
                start = pos

                while pos < len_:
                    ch = text[pos]
                    pos += 1

                    if ch == '/':
                        continue

                    if ch == '>':
                        state = State_Text
                        self.setFormat(pos - 1, 1, self.colors['tag'])
                        break

                    if not ch.isspace():
                        pos -= 1
                        state = State_AttributeName
                        break

            # at 's' in e.g. <img src=bla.png/>
            elif state == State_AttributeName:
                start = pos

                while pos < len_:
                    ch = text[pos]
                    pos += 1

                    if ch == '=':
                        state = State_AttributeValue
                        break

                    if ch in ('>', '/'):
                        state = State_InsideTag
                        break

                self.setFormat(start, pos - start, self.colors['attrname'])

            # after '=' in e.g. <img src=bla.png/>
            elif state == State_AttributeValue:
                start = pos

                # find first non-space character
                while pos < len_:
                    ch = text[pos]
                    pos += 1

                    # handle opening single quote
                    if ch == "'":
                        state = State_SingleQuote
                        self.setFormat(pos - 1, 1, self.colors['attrval'])
                        break

                    # handle opening double quote
                    if ch == '"':
                        state = State_DoubleQuote
                        self.setFormat(pos - 1, 1, self.colors['attrval'])
                        break

                    if not ch.isspace():
                        break

                if state == State_AttributeValue:
                    # attribute value without quote
                    # just stop at non-space or tag delimiter
                    start = pos
                    while pos < len_:
                        ch = text[pos]
                        if ch.isspace():
                            break
                        if ch in ('>', '/'):
                            break
                        pos += 1
                    state = State_InsideTag
                    self.setFormat(start, pos - start, self.colors['attrval'])

            # after the opening single quote in an attribute value
            elif state == State_SingleQuote:
                start = pos

                while pos < len_:
                    ch = text[pos]
                    pos += 1
                    if ch == "'":
                        break

                state = State_InsideTag

                self.setFormat(start, pos - start, self.colors['attrval'])

            # after the opening double quote in an attribute value
            elif state == State_DoubleQuote:
                start = pos

                while pos < len_:
                    ch = text[pos]
                    pos += 1
                    if ch == '"':
                        break

                state = State_InsideTag

                self.setFormat(start, pos - start, self.colors['attrval'])

            else:
                # State_Text and default
                while pos < len_:
                    ch = text[pos]
                    if ch == '<':
                        if text[pos : pos + 4] == '<!--':
                            state = State_Comment
                        elif text[pos : pos + 9].upper() == '<!DOCTYPE':
                            state = State_DocType
                        else:
                            state = State_TagStart
                        break
                    elif ch == '&':
                        start = pos
                        while pos < len_ and text[pos] != ';':
                            self.setFormat(start, pos - start, self.colors['entity'])
                            pos += 1

                    else:
                        pos += 1

        self.setCurrentBlockState(state)


# }}}


class Editor(QWidget):  # {{{
    toolbar_prefs_name = None
    data_changed = pyqtSignal()
    editor_class = EditorWidget

    def __init__(self, parent=None, one_line_toolbar=False, toolbar_prefs_name=None):
        QWidget.__init__(self, parent)
        self.toolbar_prefs_name = toolbar_prefs_name or self.toolbar_prefs_name
        self.toolbar = create_flow_toolbar(self, restrict_to_single_line=one_line_toolbar, icon_size=18)
        self.editor = self.editor_class(self)
        self.editor.data_changed.connect(self.data_changed)
        self.set_base_url = self.editor.set_base_url
        self.set_html = self.editor.set_html
        self.tabs = QTabWidget(self)
        self.tabs.setTabPosition(QTabWidget.TabPosition.South)
        self.wyswyg = QWidget(self.tabs)
        self.code_edit = PlainTextEdit(self.tabs)
        self.code_edit.setTabChangesFocus(True)
        self.source_dirty = False
        self.wyswyg_dirty = True

        self._layout = QVBoxLayout(self)
        l = QVBoxLayout(self.wyswyg)
        self.setLayout(self._layout)
        l.setContentsMargins(0, 0, 0, 0)

        l.addWidget(self.toolbar)
        l.addWidget(self.editor)
        self._layout.addWidget(self.tabs)
        self.tabs.addTab(self.wyswyg, _('&Normal view'))
        self.tabs.addTab(self.code_edit, _('&HTML source'))
        self.tabs.currentChanged[int].connect(self.change_tab)
        self.highlighter = Highlighter(self.code_edit.document())
        layout = self.layout()
        assert layout is not None
        layout.setContentsMargins(0, 0, 0, 0)
        if self.toolbar_prefs_name is not None:
            hidden = gprefs.get(self.toolbar_prefs_name)
            if hidden:
                self.hide_toolbars()

        self.toolbar.add_action(self.editor.action_undo)
        self.toolbar.add_action(self.editor.action_redo)
        self.toolbar.add_action(self.editor.action_select_all)
        self.toolbar.add_action(self.editor.action_remove_format)
        self.toolbar.add_action(self.editor.action_clear)
        self.toolbar.add_separator()

        for x in ('copy', 'cut', 'paste'):
            ac = getattr(self.editor, 'action_' + x)
            self.toolbar.add_action(ac)

        self.toolbar.add_separator()
        self.toolbar.add_action(self.editor.action_background)
        self.toolbar.add_action(self.editor.action_color)
        self.toolbar.add_separator()

        for x in ('', 'un'):
            ac = getattr(self.editor, f'action_{x}ordered_list')
            self.toolbar.add_action(ac)
        self.toolbar.add_separator()
        for x in ('superscript', 'subscript', 'indent', 'outdent'):
            self.toolbar.add_action(getattr(self.editor, 'action_' + x))
            if x in ('subscript', 'outdent'):
                self.toolbar.add_separator()

        self.toolbar.add_action(self.editor.action_block_style, popup_mode=QToolButton.ToolButtonPopupMode.InstantPopup)
        self.toolbar.add_action(self.editor.action_insert_link)
        if self.editor.insert_images_separately:
            self.toolbar.add_action(self.editor.action_insert_image)
        self.toolbar.add_action(self.editor.action_insert_hr)
        self.toolbar.add_separator()

        for x in ('bold', 'italic', 'underline', 'strikethrough'):
            ac = getattr(self.editor, 'action_' + x)
            self.toolbar.add_action(ac)
            self.addAction(ac)
        self.toolbar.add_separator()

        for x in ('left', 'center', 'right', 'justified'):
            ac = getattr(self.editor, 'action_align_' + x)
            self.toolbar.add_action(ac)
        self.toolbar.add_separator()
        QTimer.singleShot(0, self.toolbar.updateGeometry)

        self.code_edit.textChanged.connect(self.code_dirtied)
        self.editor.data_changed.connect(self.wyswyg_dirtied)

    def set_minimum_height_for_editor(self, val):
        self.editor.setMinimumHeight(val)

    @property
    def html(self):
        self.tabs.setCurrentIndex(0)
        return self.editor.html

    @html.setter
    def html(self, v):
        self.editor.html = v

    def change_tab(self, index):
        # print('reloading:', (index and self.wyswyg_dirty) or (not index and self.source_dirty))
        if index == 1:  # changing to code view
            if self.wyswyg_dirty:
                self.code_edit.setPlainText(self.editor.html)
                self.wyswyg_dirty = False
        elif index == 0:  # changing to wyswyg
            if self.source_dirty:
                self.editor.html = to_plain_text(self.code_edit)
                self.source_dirty = False

    @property
    def tab(self):
        return 'code' if self.tabs.currentWidget() is self.code_edit else 'wyswyg'

    @tab.setter
    def tab(self, val):
        self.tabs.setCurrentWidget(self.code_edit if val == 'code' else self.wyswyg)

    def wyswyg_dirtied(self, *args):
        self.wyswyg_dirty = True

    def code_dirtied(self, *args):
        self.source_dirty = True

    def hide_toolbars(self):
        self.toolbar.setVisible(False)

    def show_toolbars(self):
        self.toolbar.setVisible(True)
        QTimer.singleShot(0, self.toolbar.updateGeometry)

    def toggle_toolbars(self):
        visible = self.toolbars_visible
        getattr(self, ('hide' if visible else 'show') + '_toolbars')()
        if self.toolbar_prefs_name is not None:
            gprefs.set(self.toolbar_prefs_name, visible)

    @property
    def toolbars_visible(self):
        return self.toolbar.isVisible()

    @toolbars_visible.setter
    def toolbars_visible(self, val):
        getattr(self, ('show' if val else 'hide') + '_toolbars')()

    def set_readonly(self, what):
        self.editor.set_readonly(what)

    def hide_tabs(self):
        tab_bar = self.tabs.tabBar()
        assert tab_bar is not None
        tab_bar.setVisible(False)

    def smarten_punctuation(self):
        from calibre.ebooks.conversion.preprocess import smarten_punctuation

        html = self.html
        newhtml = smarten_punctuation(html)
        if html != newhtml:
            self.html = newhtml


# }}}

if __name__ == '__main__':
    from calibre.gui2 import Application

    app = Application([])
    w = Editor(one_line_toolbar=False)
    w.set_base_url(QUrl.fromLocalFile(os.getcwd()))
    w.resize(800, 600)
    w.setWindowFlag(Qt.WindowType.Dialog)
    w.show()
    w.html = '''<h1>Test Heading</h1><blockquote>Test blockquote</blockquote><p><span style="background-color: rgb(0, 255, 255); ">He hadn't
    set <u>out</u> to have an <em>affair</em>, <span style="font-style:italic; background-color:red">
    much</span> less a <s>long-term</s>, <b>devoted</b> one.</span><p>hello'''
    i = 'file:///home/kovid/work/calibre/resources/images/'
    w.html = f'<p>Testing <img src="{i}/donate.png"> img and another <img src="{i}/lt.png">file</p>'
    w.html = '<div><p id="moo" align="justify">Testing&nbsp;<em>a</em> link.</p><p align="justify">&nbsp;</p><p align="justify">ss</p></div>'
    app.exec()
    print(repr(w.html))



#!/usr/bin/env python
# License: GPLv3 Copyright: 2008, Kovid Goyal kovid@kovidgoyal.net

from qt.core import QCheckBox, QDialog, QDialogButtonBox, QHBoxLayout, QIcon, QLabel, QMenu, Qt, QVBoxLayout

from calibre import confirm_config_name
from calibre.gui2 import dynamic
from calibre.gui2.dialogs.message_box import Icon
from calibre.utils.localization import _


class Dialog(QDialog):
    def __init__(
        self,
        msg,
        name,
        parent,
        config_set=dynamic,
        icon='dialog_warning.png',
        title=None,
        confirm_msg=None,
        show_cancel_button=True,
        extra_button=None,
        extra_button_choices=None,
    ):
        QDialog.__init__(self, parent)
        self.setWindowTitle(title or _('Are you sure?'))
        self.setWindowIcon(QIcon.ic(icon))
        self.l = l = QVBoxLayout(self)
        self.h = h = QHBoxLayout()
        l.addLayout(h)

        self.icon_widget = Icon(self)
        self.icon_widget.set_icon(QIcon.ic(icon))

        self.msg = m = QLabel(self)
        m.setOpenExternalLinks(True)
        m.setMinimumWidth(350), m.setWordWrap(True), m.setObjectName('msg')
        m.setMaximumHeight(400)
        m.setText(msg)

        h.addWidget(self.icon_widget), h.addSpacing(10), h.addWidget(m)

        self.again = a = QCheckBox((confirm_msg or _('&Show this warning again')), self)
        a.setChecked(True), a.setObjectName('again')
        a.stateChanged.connect(self.toggle)
        l.addWidget(a)

        if show_cancel_button:
            buttons = QDialogButtonBox.StandardButton.Yes | QDialogButtonBox.StandardButton.No
            standard_button = QDialogButtonBox.StandardButton.Yes
        else:
            buttons = QDialogButtonBox.StandardButton.Ok
            standard_button = QDialogButtonBox.StandardButton.Ok
        self.buttonBox = bb = QDialogButtonBox(buttons, self)
        bb.setObjectName('buttonBox')
        bb.setFocus(Qt.FocusReason.OtherFocusReason)
        bb.accepted.connect(self.accept), bb.rejected.connect(self.reject)
        self.extra_button_clicked = False
        self.extra_button_choice = None
        if extra_button:
            b = bb.addButton(extra_button, QDialogButtonBox.ButtonRole.AcceptRole)
            if extra_button_choices:
                m = QMenu()
                b.setMenu(m)
                for acname, text in extra_button_choices.items():
                    ac = m.addAction(text)
                    ac.setObjectName(acname)
                    ac.triggered.connect(self.on_extra_choice_click)
            else:
                b.clicked.connect(self.on_extra_button_click)
        l.addWidget(bb)

        self.name = name
        self.config_set = config_set

        self.resize(self.sizeHint())
        standard_btn = bb.button(standard_button)
        assert standard_btn is not None
        standard_btn.setFocus(Qt.FocusReason.OtherFocusReason)

    def on_extra_button_click(self):
        self.extra_button_clicked = True

    def on_extra_choice_click(self):
        ac = self.sender()
        assert ac is not None
        self.extra_button_choice = ac.objectName()
        self.accept()

    def toggle(self, *args):
        self.config_set[confirm_config_name(self.name)] = self.again.isChecked()


def confirm(
    msg,
    name,
    parent=None,
    pixmap='dialog_warning.png',
    title=None,
    show_cancel_button=True,
    confirm_msg=None,
    config_set=None,
    extra_button=None,
    extra_button_choices=None,
):
    config_set = config_set or dynamic
    if not config_set.get(confirm_config_name(name), True):
        if extra_button:
            if extra_button_choices:
                return True, None
            return True, False
        return True
    d = Dialog(
        msg,
        name,
        parent,
        config_set=config_set,
        icon=pixmap,
        extra_button=extra_button,
        title=title,
        confirm_msg=confirm_msg,
        show_cancel_button=show_cancel_button,
        extra_button_choices=extra_button_choices,
    )
    ret = d.exec() == QDialog.DialogCode.Accepted
    if extra_button:
        if extra_button_choices:
            ret = ret, d.extra_button_choice
        else:
            ret = ret, d.extra_button_clicked
    return ret



#!/usr/bin/env python
# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>

"""
BASH completion for calibre commands that are too complex for simple
completion.
"""

import glob
import os
import re
import shlex
import sys

prints = print


def split(src):
    try:
        return shlex.split(src)
    except ValueError:
        try:
            return shlex.split(src + '"')
        except ValueError:
            return shlex.split(src + "'")


def files_and_dirs(prefix, allowed_exts=[]):
    prefix = os.path.expanduser(prefix)
    for i in glob.iglob(prefix + '*'):
        _, ext = os.path.splitext(i)
        ext = ext.lower().replace('.', '')
        if os.path.isdir(i):
            yield i + os.sep
        elif allowed_exts is None or ext in allowed_exts:
            yield i + ' '


def get_opts_from_parser(parser, prefix):
    def do_opt(opt):
        for x in opt._long_opts:
            if x.startswith(prefix):
                yield x
        for x in opt._short_opts:
            if x.startswith(prefix):
                yield x

    for o in parser.option_list:
        for x in do_opt(o):
            yield x + ' '
    for g in parser.option_groups:
        for o in g.option_list:
            for x in do_opt(o):
                yield x + ' '


def send(ans):
    pat = re.compile(r'([^0-9a-zA-Z_./-])')
    for x in sorted(set(ans)):
        x = pat.sub(lambda m: '\\' + m.group(1), x)
        if x.endswith('\\ '):
            x = x[:-2] + ' '
        prints(x)


class EbookConvert:
    def __init__(self, comp_line, pos):
        words = split(comp_line[:pos])
        char_before = comp_line[pos - 1]
        prefix = words[-1] if words[-1].endswith(char_before) else ''
        wc = len(words)
        if not prefix:
            wc += 1
        self.words = words
        self.prefix = prefix
        self.previous = words[-2 if prefix else -1]
        from calibre.utils.serialize import msgpack_loads

        self.cache = msgpack_loads(
            open(os.path.join(getattr(sys, 'resources_location'), 'ebook-convert-complete.calibre_msgpack'), 'rb').read(),
            use_list=False,
        )
        self.complete(wc)

    def complete(self, wc):
        if wc == 2:
            self.complete_input()
        elif wc == 3:
            self.complete_output()
        else:
            q = list(self.words[1:3])
            q = [os.path.splitext(x)[0 if x.startswith('.') else 1].partition('.')[-1].lower() for x in q]
            if not q[1]:
                q[1] = 'oeb'
            q = tuple(q)
            if q in self.cache:
                ans = [x for x in self.cache[q] if x.startswith(self.prefix)]
            else:
                from calibre.ebooks.conversion.cli import create_option_parser
                from calibre.utils.logging import Log

                log = Log()
                log.outputs = []
                ans = []
                if not self.prefix or self.prefix.startswith('-'):
                    try:
                        parser, _ = create_option_parser(self.words[:3], log)
                        ans += list(get_opts_from_parser(parser, self.prefix))
                    except Exception:
                        pass
            if self.previous.startswith('-'):
                ans += list(files_and_dirs(self.prefix, None))
            send(ans)

    def complete_input(self):
        ans = list(files_and_dirs(self.prefix, self.cache['input_fmts']))
        ans += [t for t in self.cache['input_recipes'] if t.startswith(self.prefix)]
        send(ans)

    def complete_output(self):
        fmts = self.cache['output']
        ans = list(files_and_dirs(self.prefix, fmts))
        ans += ['.' + x + ' ' for x in fmts if ('.' + x).startswith(self.prefix)]
        send(ans)


def main(args=sys.argv):
    comp_line, pos = os.environ['COMP_LINE'], int(os.environ['COMP_POINT'])
    module = split(comp_line)[0].split(os.sep)[-1]
    if module == 'ebook-convert':
        EbookConvert(comp_line, pos)

    return 0


if __name__ == '__main__':
    raise sys.exit(main())



# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid at kovidgoyal.net>

import textwrap

from qt.core import QCheckBox, QComboBox, QLabel, QLineEdit, QListWidgetItem, Qt, QWidget

from calibre.ebooks import BOOK_EXTENSIONS
from calibre.gui2 import error_dialog, question_dialog
from calibre.gui2.device_drivers.configwidget_ui import Ui_ConfigWidget
from calibre.utils.formatter import validation_formatter
from calibre.utils.localization import _


class ConfigWidget(QWidget, Ui_ConfigWidget):
    def __init__(
        self,
        settings,
        all_formats,
        supports_subdirs,
        must_read_metadata,
        supports_use_author_sort,
        extra_customization_message,
        device,
        extra_customization_choices=None,
    ):

        QWidget.__init__(self)
        Ui_ConfigWidget.__init__(self)
        self.setupUi(self)

        self.settings = settings

        all_formats = set(all_formats)
        self.calibre_known_formats = device.FORMATS
        try:
            self.device_name = device.get_gui_name()
        except TypeError:
            self.device_name = getattr(device, 'gui_name', None) or _('Device')
        if device.USER_CAN_ADD_NEW_FORMATS:
            all_formats = all_formats | set(BOOK_EXTENSIONS)

        format_map = settings.format_map
        disabled_formats = all_formats.difference(format_map)
        for format in format_map + sorted(disabled_formats):
            item = QListWidgetItem(format, self.columns)
            item.setData(Qt.ItemDataRole.UserRole, (format))
            item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsSelectable)
            item.setCheckState(Qt.CheckState.Checked if format in format_map else Qt.CheckState.Unchecked)

        self.column_up.clicked.connect(self.up_column)
        self.column_down.clicked.connect(self.down_column)

        if device.HIDE_FORMATS_CONFIG_BOX:
            self.groupBox.hide()

        if supports_subdirs:
            self.opt_use_subdirs.setChecked(self.settings.use_subdirs)
        else:
            self.opt_use_subdirs.hide()
        if not must_read_metadata:
            self.opt_read_metadata.setChecked(self.settings.read_metadata)
        else:
            self.opt_read_metadata.hide()
        if supports_use_author_sort:
            self.opt_use_author_sort.setChecked(self.settings.use_author_sort)
        else:
            self.opt_use_author_sort.hide()
        if extra_customization_message:
            extra_customization_choices = extra_customization_choices or {}

            def parse_msg(m):
                msg, _, tt = m.partition(':::') if m else ('', '', '')
                return msg.strip(), textwrap.fill(tt.strip(), 100)

            if isinstance(extra_customization_message, list):
                self.opt_extra_customization = []
                if len(extra_customization_message) > 6:

                    def row_func(x, y):
                        return x // 2 * 2 + y

                    def col_func(x):
                        return x % 2

                else:

                    def row_func(x, y):
                        return x * 2 + y

                    def col_func(x):
                        return 0

                for i, m in enumerate(extra_customization_message):
                    label_text, tt = parse_msg(m)
                    if not label_text:
                        self.opt_extra_customization.append(None)
                        continue
                    if isinstance(settings.extra_customization[i], bool):
                        opt_ec = QCheckBox(label_text)
                        self.opt_extra_customization.append(opt_ec)
                        opt_ec.setToolTip(tt)
                        opt_ec.setChecked(bool(settings.extra_customization[i]))
                    elif i in extra_customization_choices:
                        cb = QComboBox(self)
                        self.opt_extra_customization.append(cb)
                        l = QLabel(label_text)
                        l.setToolTip(tt), cb.setToolTip(tt), l.setBuddy(cb), cb.setToolTip(tt)
                        for li in sorted(extra_customization_choices[i]):
                            cb.addItem(li)
                        cb.setCurrentIndex(max(0, cb.findText(settings.extra_customization[i])))
                    else:
                        opt_ec_i = QLineEdit(self)
                        self.opt_extra_customization.append(opt_ec_i)
                        l = QLabel(label_text)
                        l.setToolTip(tt)
                        opt_ec_i.setToolTip(tt)
                        l.setBuddy(opt_ec_i)
                        l.setWordWrap(True)
                        opt_ec_i.setText(settings.extra_customization[i])
                        opt_ec_i.setCursorPosition(0)
                        self.extra_layout.addWidget(l, row_func(i, 0), col_func(i))
                    self.extra_layout.addWidget(self.opt_extra_customization[i], row_func(i, 1), col_func(i))
            else:
                self.opt_extra_customization = QLineEdit()
                label_text, tt = parse_msg(extra_customization_message)
                l = QLabel(label_text)
                l.setToolTip(tt)
                l.setBuddy(self.opt_extra_customization)
                l.setWordWrap(True)
                if settings.extra_customization:
                    self.opt_extra_customization.setText(settings.extra_customization)
                    self.opt_extra_customization.setCursorPosition(0)
                self.opt_extra_customization.setCursorPosition(0)
                self.extra_layout.addWidget(l, 0, 0)
                self.extra_layout.addWidget(self.opt_extra_customization, 1, 0)
        self.opt_save_template.setText(settings.save_template)

    def up_column(self):
        idx = self.columns.currentRow()
        if idx > 0:
            self.columns.insertItem(idx - 1, self.columns.takeItem(idx))
            self.columns.setCurrentRow(idx - 1)

    def down_column(self):
        idx = self.columns.currentRow()
        if idx < self.columns.count() - 1:
            self.columns.insertItem(idx + 1, self.columns.takeItem(idx))
            self.columns.setCurrentRow(idx + 1)

    def format_map(self):
        formats = []
        for i in range(self.columns.count()):
            col_item = self.columns.item(i)
            assert col_item is not None
            if col_item.checkState() == Qt.CheckState.Checked:
                formats.append(str(col_item.data(Qt.ItemDataRole.UserRole) or ''))
        return formats

    def use_subdirs(self):
        return self.opt_use_subdirs.isChecked()

    def read_metadata(self):
        return self.opt_read_metadata.isChecked()

    def use_author_sort(self):
        return self.opt_use_author_sort.isChecked()

    def validate(self):
        formats = set(self.format_map())
        extra = formats - set(self.calibre_known_formats)
        if extra:
            fmts = sorted(x.upper() for x in extra)
            if not question_dialog(
                self,
                _('Unknown formats'),
                _(
                    'You have enabled the <b>{0}</b> formats for'
                    ' your {1}. The {1} may not support them.'
                    ' If you send these formats to your {1} they '
                    'may not work. Are you sure?'
                ).format((', '.join(fmts)), self.device_name),
            ):
                return False

        tmpl = str(self.opt_save_template.text())
        try:
            validation_formatter.validate(tmpl)
            return True
        except Exception as err:
            error_dialog(
                self,
                _('Invalid template'),
                '<p>' + _('The template %s is invalid:') % tmpl + '<br>' + str(err),
                show=True,
            )

            return False



#!/usr/bin/env python
# License: GPLv3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>

import functools
import importlib
import importlib.resources
import json
import os
import queue
import sys
import tempfile
import threading
import unittest

from calibre.constants import debug, isbsd, islinux, ismacos, iswindows
from calibre.utils.monotonic import monotonic
from calibre.utils.serialize import msgpack_dumps
from polyglot.binary import as_hex_unicode

is_ci = os.environ.get('CI', '').lower() == 'true'

# Below this many tests, skip worker processes and run in-process instead.
_PARALLEL_MIN = 32

# ANSI colours – only when stdout is a real terminal.
_IS_TTY: bool = sys.stdout.isatty()
_USE_COLOR: bool = _IS_TTY
_R = '\033[0m'  # reset
_BD = '\033[1m'  # bold
_DM = '\033[2m'  # dim
_GR = '\033[32m'  # green
_RD = '\033[31m'  # red
_YL = '\033[33m'  # yellow
_CY = '\033[36m'  # cyan


def _c(code: str, text: str) -> str:
    """Wrap *text* in an ANSI escape sequence if colour output is enabled."""
    return f'{code}{text}{_R}' if _USE_COLOR else text


def no_endl(f):
    @functools.wraps(f)
    def func(*args, **kwargs):
        self = f.__self__
        orig = self.stream.writeln
        self.stream.writeln = self.stream.write
        try:
            return f(*args, **kwargs)
        finally:
            self.stream.writeln = orig

    return func


class TestResult(unittest.TextTestResult):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_time = {}
        for x in ('Success', 'Error', 'Failure', 'Skip', 'ExpectedFailure', 'UnexpectedSuccess'):
            x = 'add' + x
            setattr(self, x, no_endl(getattr(self, x)))
        self.times = {}

    def startTest(self, test):
        self.start_time[test] = monotonic()
        return super().startTest(test)

    def stopTest(self, test):
        orig = self.stream.writeln
        try:
            self.stream.writeln = self.stream.write  # type: ignore
            super().stopTest(test)
            elapsed = monotonic()
            elapsed -= self.start_time.get(test, elapsed)
            self.times[test] = elapsed
        finally:
            self.stream.writeln = orig  # type: ignore
            self.stream.writeln(f' [{elapsed:.1f} s]')

    def stopTestRun(self):
        super().stopTestRun()
        if self.wasSuccessful():
            tests = sorted(self.times, key=self.times.__getitem__, reverse=True)
            slowest = [f'{t.id()} [{self.times[t]:.1f} s]' for t in tests[:3]]
            if len(slowest) > 1:
                self.stream.writeln('\nSlowest tests: {}'.format(' '.join(slowest)))


def find_tests_in_package(package, excludes=('main.py',)):
    items = [path.name for path in importlib.resources.files(package).iterdir()]
    suits = []
    excludes = set(excludes) | {x + 'c' for x in excludes}
    seen = set()
    for x in items:
        if (x.endswith(('.py', '.pyc'))) and x not in excludes:
            q = x.rpartition('.')[0]
            if q in seen:
                continue
            seen.add(q)
            m = importlib.import_module(package + '.' + x.partition('.')[0])
            suits.append(unittest.defaultTestLoader.loadTestsFromModule(m))
    return unittest.TestSuite(suits)


def itertests(suite):
    stack = [suite]
    while stack:
        suite = stack.pop()
        for test in suite:
            if isinstance(test, unittest.TestSuite):
                stack.append(test)
                continue
            if test.__class__.__name__ == 'ModuleImportFailure':
                raise Exception(f'Failed to import a test module: {test}')
            yield test


def init_env():
    from calibre.ebooks.metadata.book.base import reset_field_metadata
    from calibre.ebooks.oeb.polish.utils import setup_css_parser_serialization
    from calibre.utils.config_base import reset_tweaks_to_default

    reset_tweaks_to_default()
    reset_field_metadata()
    setup_css_parser_serialization()


def filter_tests(suite, test_ok):
    ans = unittest.TestSuite()
    added = set()
    for test in itertests(suite):
        if test_ok(test) and test not in added:
            ans.addTest(test)
            added.add(test)
    return ans


def filter_tests_by_name(suite, *names):
    names = {x if x.startswith('test_') else 'test_' + x for x in names}

    def q(test):
        return test._testMethodName in names

    return filter_tests(suite, q)


def remove_tests_by_name(suite, *names):
    names = {x if x.startswith('test_') else 'test_' + x for x in names}

    def q(test):
        return test._testMethodName not in names

    return filter_tests(suite, q)


def filter_tests_by_module(suite, *names):
    names = frozenset(names)

    def q(test):
        m = test.__class__.__module__.rpartition('.')[-1]
        return m in names

    return filter_tests(suite, q)


# ─── Parallel runner ──────────────────────────────────────────────────────────


class PipeTestResult(unittest.TestResult):
    """
    Writes test events as newline-delimited JSON to a text stream.
    Used inside worker processes; the master reads and aggregates the events.

    Event schema (field 'e'):
      's'    – test started (no elapsed)
      'ok'   – addSuccess
      'fail' – addFailure   (+ 'msg')
      'err'  – addError     (+ 'msg')
      'skip' – addSkip      (+ 'msg')
      'xf'   – addExpectedFailure (+ 'msg')
      'xp'   – addUnexpectedSuccess
    All events except 's' carry 't' (elapsed seconds as float).
    """

    def __init__(self, write_fd: int) -> None:
        super().__init__()
        self._out = open(write_fd, 'w', buffering=1, closefd=True)
        self._start: dict[str, float] = {}

    def _emit(self, record: dict) -> None:
        try:
            self._out.write(json.dumps(record, ensure_ascii=False) + '\n')
            self._out.flush()
        except OSError:
            pass

    def startTest(self, test: unittest.TestCase) -> None:
        super().startTest(test)
        self._start[test.id()] = monotonic()
        self._emit({'e': 's', 'id': test.id()})

    def _elapsed(self, test: unittest.TestCase) -> float:
        return monotonic() - self._start.pop(test.id(), monotonic())

    def addSuccess(self, test: unittest.TestCase) -> None:
        super().addSuccess(test)
        self._emit({'e': 'ok', 'id': test.id(), 't': self._elapsed(test)})

    def addFailure(self, test: unittest.TestCase, err) -> None:
        super().addFailure(test, err)
        self._emit({'e': 'fail', 'id': test.id(), 't': self._elapsed(test), 'msg': self._exc_info_to_string(err, test)})  # type: ignore

    def addError(self, test: unittest.TestCase, err) -> None:
        super().addError(test, err)
        self._emit({'e': 'err', 'id': test.id(), 't': self._elapsed(test), 'msg': self._exc_info_to_string(err, test)})  # type: ignore

    def addSkip(self, test: unittest.TestCase, reason: str) -> None:
        super().addSkip(test, reason)
        self._emit({'e': 'skip', 'id': test.id(), 't': self._elapsed(test), 'msg': reason})

    def addExpectedFailure(self, test: unittest.TestCase, err) -> None:
        super().addExpectedFailure(test, err)
        self._emit({'e': 'xf', 'id': test.id(), 't': self._elapsed(test), 'msg': self._exc_info_to_string(err, test)})  # type: ignore

    def addUnexpectedSuccess(self, test: unittest.TestCase) -> None:
        super().addUnexpectedSuccess(test)
        self._emit({'e': 'xp', 'id': test.id(), 't': self._elapsed(test)})

    def close(self) -> None:
        self._out.close()


def _worker_entry(write_fd_or_handle: int) -> None:
    """
    Entry point run inside subprocess workers.
    On POSIX, write_fd_or_handle is the write-end file descriptor number.
    On Windows, it is the underlying Windows HANDLE value (since POSIX fd
    numbers are per-process and are not inherited by number in child processes).
    """
    if iswindows:
        import msvcrt

        write_fd = msvcrt.open_osfhandle(write_fd_or_handle, os.O_WRONLY)
    else:
        write_fd = write_fd_or_handle
    os.set_inheritable(write_fd, False)
    debug(False)
    inp = sys.stdin.read()
    test_ids = frozenset(json.loads(inp))
    from calibre.utils.run_tests import PipeTestResult, filter_tests, find_tests, init_env

    init_env()
    suite = filter_tests(find_tests(), lambda t: t.id() in test_ids)
    result = PipeTestResult(write_fd)
    suite.run(result)
    result.close()


def _chunk_round_robin(tests: list, num_workers: int) -> list[list]:
    """Distribute tests evenly across workers using round-robin assignment."""
    chunks: list[list] = [[] for _ in range(num_workers)]
    for i, t in enumerate(tests):
        chunks[i % num_workers].append(t)
    return [c for c in chunks if c]


def _start_worker(idx: int, chunk: list, worker_cmd: list[str] | None = None) -> tuple:
    """
    Start a subprocess worker for the given test chunk.
    Returns (Popen, read_pipe_file).

    When *worker_cmd* is provided it must be a list whose last element is the
    placeholder to which the Python code string will be appended, e.g.
    ``[sys.executable, 'setup.py', 'test', '--worker']``.  The code string is
    appended as a new argument so the worker runs via setup.py instead of
    calibre-debug.  When *worker_cmd* is None (the default) the traditional
    ``calibre-debug -c <code>`` invocation is used.

    On POSIX, the write-end file descriptor number is passed directly to the
    worker (close_fds=False lets it be inherited).  On Windows, POSIX fd
    numbers are per-process and cannot be inherited by number; instead we pass
    the underlying Windows HANDLE value and the worker reconstructs a file
    descriptor from it via msvcrt.open_osfhandle.  Workers are always spawned
    one at a time in a loop and the parent closes its copy of the write-end
    before creating the next pipe, so no worker accidentally inherits another
    worker's write-end.
    """
    import subprocess

    r, w = os.pipe()
    os.set_inheritable(r, False)
    os.set_inheritable(w, True)

    test_ids_json = json.dumps([t.id() for t in chunk])

    if iswindows:
        import msvcrt

        w_arg = msvcrt.get_osfhandle(w)
        code = f'from calibre.utils.run_tests import _worker_entry; _worker_entry({w_arg})'
    else:
        code = f'from calibre.utils.run_tests import _worker_entry; _worker_entry({w})'

    if worker_cmd is not None:
        cmd = worker_cmd + [code]
    else:
        from calibre.startup import get_debug_executable

        cmd = get_debug_executable() + ['-c', code]

    with tempfile.NamedTemporaryFile(delete=False) as output_tf:
        env = os.environ.copy()
        from calibre.ptempfile import base_dir

        tdir = os.path.join(base_dir(), f'w{idx}')
        os.mkdir(tdir)
        env['CALIBRE_WORKER_TEMP_DIR'] = as_hex_unicode(msgpack_dumps(tdir))
        proc = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=output_tf,
            stderr=output_tf,
            close_fds=False,
            env=env,
        )
    assert proc.stdin is not None
    proc.stdin.write(test_ids_json.encode())
    proc.stdin.close()
    os.close(w)
    return proc, os.fdopen(r, 'r', errors='replace'), output_tf.name


_STATUS_INTERVAL = 1.0  # seconds between status lines when stdout is not a TTY
_last_status_print: list[float] = [0.0]


def _print_status(completed: int, total: int, start: float, n_bad: int = 0) -> None:
    elapsed = monotonic() - start
    if _IS_TTY:
        time_part = _c(_DM, f'[{elapsed:6.1f}s]')
        count_part = _c(_RD if n_bad else _BD, f'{completed}/{total}')
        sys.stdout.write(f'\r{time_part} {count_part} tests completed   ')
        sys.stdout.flush()
    else:
        now = monotonic()
        if now - _last_status_print[0] >= _STATUS_INTERVAL:
            _last_status_print[0] = now
            print(f'[{elapsed:6.1f}s] {completed}/{total} tests completed', flush=True)


def _print_summary(
    failures: list[tuple[str, str]],
    errors: list[tuple[str, str]],
    skips: list[tuple[str, str]],
    ok_count: int,
    xfail_count: int,
    times: dict[str, float],
    total: int,
    elapsed: float,
    lost_ids: list[str],
) -> None:
    """Print the failure details, one-line stats, and five slowest tests."""
    for label, colour, items in (
        ('FAIL', _RD, failures),
        ('ERROR', _BD + _RD, errors),
    ):
        for tid, msg in items:
            print(f'\n{_c(_BD, "=" * 70)}\n{_c(colour, label + ": " + tid)}\n{_c(_DM, "-" * 70)}')
            print(msg)

    def _stat(label: str, count: int, colour: str) -> str:
        s = f'{label}: {count}'
        return _c(colour, s) if count else s

    lost = len(lost_ids)
    print(f'\n{_c(_BD, f"Ran {total} tests")} in {elapsed:.1f}s')
    parts = [
        _c(_GR, f'OK: {ok_count}'),
        _stat('FAIL', len(failures), _RD),
        _stat('ERROR', len(errors), _RD),
        _stat('SKIP', len(skips), _YL),
    ]
    if xfail_count:
        parts.append(_c(_DM, f'XFAIL/XPASS: {xfail_count}'))
    if lost:
        parts.append(_c(_RD, f'LOST (worker crash?): {lost}'))
    print('  '.join(parts))
    if lost_ids:
        print(_c(_RD, '\nLost tests (never reported back):'))
        for tid in lost_ids:
            print(f'  {_c(_RD, tid)}')

    slowest = sorted(times.items(), key=lambda kv: kv[1], reverse=True)[:5]
    if slowest:
        print(_c(_DM, '\nSlowest tests:'))
        for tid, t in slowest:
            print(f'  {_c(_DM, f"{t:6.1f}s")}  {tid}')


class TooFewTests(Exception):
    pass


def run_parallel(suite: unittest.TestSuite, num_workers: int = 0, worker_cmd: list[str] | None = None) -> int:
    """
    Run *suite* spread across independent worker processes.
    Each process receives a round-robin slice of the tests and streams results
    back over a pipe as newline-delimited JSON.  The master prints a live status
    line and, once all workers finish, prints failures/errors and a summary.

    When the suite is smaller than *_PARALLEL_MIN* tests the overhead of
    spawning workers outweighs the benefit; in that case the suite is run
    in the current process instead.

    Returns the number of failures + errors (0 means all passed).
    """

    all_tests = sorted(itertests(suite), key=lambda t: t.id())
    total = len(all_tests)
    if not total:
        print('No tests found.', file=sys.stderr)
        return 0

    if num_workers <= 0:
        num_workers = max(1, min(total, os.cpu_count() or 4))
    num_workers = min(num_workers, total)

    if total < _PARALLEL_MIN or num_workers < 2:
        raise TooFewTests()
    chunks = _chunk_round_robin(all_tests, num_workers)

    pipes: list = []
    workers: list[tuple] = []  # (proc, output_file)
    for i, chunk in enumerate(chunks):
        proc, pipe, output_file = _start_worker(i, chunk, worker_cmd=worker_cmd)
        workers.append((proc, output_file))
        pipes.append(pipe)

    ev_queue: queue.Queue[dict | None] = queue.Queue()

    def _reader(pipe) -> None:
        try:
            for line in pipe:
                line = line.strip()
                if line:
                    try:
                        ev_queue.put(json.loads(line))
                    except json.JSONDecodeError:
                        pass
        finally:
            pipe.close()
        ev_queue.put(None)  # sentinel: this worker's pipe is exhausted

    for pipe in pipes:
        threading.Thread(target=_reader, args=(pipe,), daemon=True).start()

    failures: list[tuple[str, str]] = []
    errors: list[tuple[str, str]] = []
    skips: list[tuple[str, str]] = []
    times: dict[str, float] = {}
    ok_count = 0
    xfail_count = 0
    completed = 0
    done_workers = 0
    start_time = monotonic()

    while done_workers < len(pipes):
        try:
            event = ev_queue.get(timeout=0.25)
        except queue.Empty:
            _print_status(completed, total, start_time, len(failures) + len(errors))
            continue

        if event is None:
            done_workers += 1
        else:
            e = event.get('e')
            tid = event.get('id', '')
            t = float(event.get('t', 0.0))
            if e != 's':  # skip start events, they carry no result
                completed += 1
                times[tid] = t
                if e == 'ok':
                    ok_count += 1
                elif e == 'fail':
                    failures.append((tid, event.get('msg', '')))
                elif e == 'err':
                    errors.append((tid, event.get('msg', '')))
                elif e == 'skip':
                    skips.append((tid, event.get('msg', '')))
                elif e in ('xf', 'xp'):
                    xfail_count += 1

        _print_status(completed, total, start_time, len(failures) + len(errors))

    # Reap worker processes; report crashes.
    try:
        for proc, output_path in workers:
            rc = proc.wait()
            if rc != 0:
                with open(output_path, 'rb') as f:
                    output = f.read().decode(errors='replace')
                print(f'\n{_c(_RD, f"Worker process crashed (exit code {rc})")}', flush=True)
                if output.strip():
                    print(output, flush=True)
    finally:
        for _, output_path in workers:
            try:
                os.unlink(output_path)
            except OSError:
                pass

    if _IS_TTY:
        print()

    elapsed = monotonic() - start_time
    lost_ids = [t.id() for t in all_tests if t.id() not in times]
    _print_summary(failures, errors, skips, ok_count, xfail_count, times, total, elapsed, lost_ids)
    return 1 if len(failures) + len(errors) + len(lost_ids) else 0


# ─── Test discovery ───────────────────────────────────────────────────────────


class TestImports(unittest.TestCase):
    def base_check(self, base, exclude_packages, exclude_modules):
        import importlib

        import_base = os.path.dirname(base)
        count = 0
        for root, dirs, files in os.walk(base):
            for d in tuple(dirs):
                if not os.path.isfile(os.path.join(root, d, '__init__.py')):
                    dirs.remove(d)
            for fname in files:
                module_name, ext = os.path.splitext(fname)
                if ext != '.py':
                    continue
                path = os.path.join(root, module_name)
                relpath = os.path.relpath(path, import_base).replace(os.sep, '/')
                full_module_name = '.'.join(relpath.split('/'))
                if full_module_name.endswith('.__init__'):
                    full_module_name = full_module_name.rpartition('.')[0]
                if full_module_name in exclude_modules or ('.' in full_module_name and full_module_name.rpartition('.')[0] in exclude_packages):
                    continue
                importlib.import_module(full_module_name)
                count += 1
        return count

    def test_import_of_all_python_modules(self):

        exclude_packages = {'calibre.devices.mtp.unix.upstream'}
        exclude_modules = {'calibre.web.automate.browser'}
        if not iswindows:
            exclude_modules |= {'calibre.utils.iphlpapi', 'calibre.utils.open_with.windows', 'calibre.devices.winusb'}
            exclude_packages |= {'calibre.utils.winreg', 'calibre.utils.windows'}
        if not ismacos:
            exclude_modules.add('calibre.utils.open_with.osx')
        if not islinux:
            exclude_modules |= {
                'calibre.linux',
                'calibre.gui2.tts.speechd',
                'calibre.utils.linux_trash',
                'calibre.utils.open_with.linux',
                'calibre.gui2.linux_file_dialogs',
            }
        if 'SKIP_SPEECH_TESTS' in os.environ:
            exclude_packages.add('calibre.gui2.tts')
        if not isbsd:
            exclude_modules.add('calibre.devices.usbms.hal')
        d = os.path.dirname
        SRC = d(d(d(os.path.abspath(__file__))))
        self.assertGreater(self.base_check(os.path.join(SRC, 'odf'), exclude_packages, exclude_modules), 10)
        base = os.path.join(SRC, 'calibre')
        self.assertGreater(self.base_check(base, exclude_packages, exclude_modules), 1000)

        # Virtual imports for legacy compatibility
        import calibre.web.feeds.feedparser as f  # type: ignore

        del f
        from calibre.ebooks.markdown import Markdown  # type: ignore

        del Markdown


def find_tests(which_tests=None, exclude_tests=None):
    ans = []
    a = ans.append

    def ok(x):
        return (not which_tests or x in which_tests) and (not exclude_tests or x not in exclude_tests)

    if ok('fork'):
        from calibre.utils.forked_map import find_tests

        a(find_tests())
    if ok('build'):
        from calibre.test_build import find_tests

        a(find_tests(only_build=True))
    if ok('srv'):
        from calibre.srv.tests.main import find_tests

        a(find_tests())
    if ok('db'):
        from calibre.db.tests.main import find_tests

        a(find_tests())
    if ok('polish'):
        from calibre.ebooks.oeb.polish.tests.main import find_tests

        a(find_tests())
        from calibre.ebooks.oeb.polish.tests.structure import find_tests

        a(find_tests())
    if ok('opf'):
        from calibre.ebooks.metadata.opf2 import suite

        a(suite())
        from calibre.ebooks.metadata.opf3_test import suite

        a(suite())
    if ok('css'):
        from tinycss.tests.main import find_tests

        a(find_tests())
        from calibre.ebooks.oeb.normalize_css import test_normalization

        a(test_normalization(return_tests=True))
        from calibre.ebooks.css_transform_rules import test

        a(test(return_tests=True))
        from calibre.ebooks.html_transform_rules import test

        a(test(return_tests=True))
        from css_selectors.tests import find_tests

        a(find_tests())
    if ok('docx'):
        from calibre.ebooks.docx.fields import test_parse_fields

        a(test_parse_fields(return_tests=True))
        from calibre.ebooks.docx.writer.utils import test_convert_color

        a(test_convert_color(return_tests=True))
    if ok('cfi'):
        from calibre.ebooks.epub.cfi.tests import find_tests

        a(find_tests())
    if ok('matcher'):
        from calibre.utils.matcher import test

        a(test(return_tests=True))
    if ok('scraper'):
        from calibre.scraper.test_fetch_backend import find_tests

        a(find_tests())
        from calibre.web.automate.test_worker import find_tests

        a(find_tests())
    if ok('icu'):
        from calibre.utils.icu_test import find_tests

        a(find_tests())
    if ok('smartypants'):
        from calibre.utils.smartypants import run_tests

        a(run_tests(return_tests=True))
    if ok('ebooks'):
        from calibre.ebooks.conversion.plugins.archive_input import find_tests

        a(find_tests())
        from calibre.ebooks.conversion.plugins.txt_input import find_tests

        a(find_tests())
        from calibre.ebooks.metadata.rtf import find_tests

        a(find_tests())
        from calibre.ebooks.metadata.html import find_tests

        a(find_tests())
        from calibre.utils.xml_parse import find_tests

        a(find_tests())
        from calibre.gui2.viewer.annotations import find_tests

        a(find_tests())
        from calibre.gui2.library.test_annotations import find_tests

        a(find_tests())
        from calibre.ebooks.html_entities import find_tests

        a(find_tests())
        from calibre.spell.dictionary import find_tests

        a(find_tests())
    if ok('ai'):
        from calibre.ai.utils import find_tests

        a(find_tests())
    if ok('misc'):
        from calibre.ebooks.html.input import find_tests

        a(find_tests())
        from calibre.ebooks.metadata.test_author_sort import find_tests

        a(find_tests())
        from calibre.ebooks.metadata.tag_mapper import find_tests

        a(find_tests())
        from calibre.ebooks.metadata.author_mapper import find_tests

        a(find_tests())
        from calibre.utils.shared_file import find_tests

        a(find_tests())
        from calibre.utils.test_lock import find_tests

        a(find_tests())
        from calibre.utils.search_query_parser_test import find_tests

        a(find_tests())
        from calibre.utils.html2text import find_tests

        a(find_tests())
        from calibre.utils.shm import find_tests

        a(find_tests())
        from calibre.library.comments import find_tests

        a(find_tests())
        from calibre.ebooks.compression.palmdoc import find_tests

        a(find_tests())
        from calibre.gui2.viewer.convert_book import find_tests

        a(find_tests())
        from calibre.utils.hyphenation.test_hyphenation import find_tests

        a(find_tests())
        from calibre.live import find_tests

        a(find_tests())
        from calibre.utils.copy_files_test import find_tests

        a(find_tests())
        from calibre.utils.safe_atexit import find_tests

        a(find_tests())
        from calibre.gui2.listener import find_tests

        a(find_tests())
        if iswindows:
            from calibre.utils.windows.wintest import find_tests

            a(find_tests())
        a(unittest.defaultTestLoader.loadTestsFromTestCase(TestImports))
        from calibre.utils.translator.test_translator import find_tests

        a(find_tests())
        from calibre.utils.config_base import find_tests

        a(find_tests())
        from calibre.utils.zipfile import find_tests

        a(find_tests())
    if ok('dbcli'):
        from calibre.db.cli.tests import find_tests

        a(find_tests())

    tests = unittest.TestSuite(ans)
    return tests


# ─── Public entry points ──────────────────────────────────────────────────────


def run_tests(find_tests_fn, verbosity=4):
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        'name',
        nargs='?',
        default=None,
        help='The name of the test to run, for example: writing.WritingTest.many_many_basic or .many_many_basic for a shortcut',
    )
    args = parser.parse_args()
    tests = find_tests_fn()
    if args.name:
        if args.name.startswith('.'):
            tests = filter_tests_by_name(tests, args.name[1:])
        else:
            tests = filter_tests_by_module(tests, args.name)
        if not tests._tests:
            raise SystemExit(f'No test named {args.name} found')
    run_cli(tests, verbosity, buffer=not args.name)


def run_test(test_name, verbosity=4):
    # calibre-debug -t test_name
    which_tests = None
    if test_name.startswith('@'):
        which_tests = (test_name[1:],)
    tests = find_tests(which_tests)
    buffer = True
    if test_name != 'all':
        buffer = False
        if test_name.startswith('.'):
            tests = filter_tests_by_module(tests, test_name[1:])
        elif test_name.startswith('@'):
            pass
        else:
            tests = filter_tests_by_name(tests, test_name)
    if not tests._tests:
        raise SystemExit(f'No test named {test_name} found')
    run_cli(tests, verbosity, buffer=buffer)


def run_cli(suite, verbosity=4, buffer=True, worker_cmd: list[str] | None = None):
    """
    Run *suite* and exit.
    When *buffer* is True (the default, used for full-suite runs) tests are
    distributed across worker processes via run_parallel().
    When *buffer* is False (used for single named-test runs) the suite runs
    single-threaded with verbose TextTestRunner output.
    *worker_cmd*, when provided, is forwarded to run_parallel() / _start_worker()
    as the command prefix used to spawn worker processes.
    """
    init_env()
    if buffer:
        try:
            raise SystemExit(run_parallel(suite, worker_cmd=worker_cmd))
        except TooFewTests:
            pass
    r = unittest.TextTestRunner
    r.resultclass = unittest.TextTestResult if verbosity < 2 else TestResult
    result = r(verbosity=verbosity, buffer=False).run(suite)
    rc = 0 if result.wasSuccessful() else 1
    raise SystemExit(rc)



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

import os

from calibre.customize.conversion import InputFormatPlugin, OptionRecommendation
from calibre.utils.localization import _
from polyglot.builtins import as_bytes

ENGINES = 'calibre', 'pdftohtml'


class PDFInput(InputFormatPlugin):
    name = 'PDF Input'
    author = 'Kovid Goyal and John Schember and Alan Pettigrew'
    description = _('Convert PDF files to HTML')
    file_types = {'pdf'}
    commit_name = 'pdf_input'

    options = {
        OptionRecommendation(name='no_images', recommended_value=False, help=_('Do not extract images from the document')),
        OptionRecommendation(
            name='unwrap_factor',
            recommended_value=0.45,
            help=_(
                'Scale used to determine the length at which a line should '
                'be unwrapped. Valid values are a decimal between 0 and 1. The '
                'default is 0.45, just below the median line length.'
            ),
        ),
        OptionRecommendation(
            name='pdf_engine',
            recommended_value='calibre',
            choices=('calibre', 'pdftohtml'),
            help=_('The PDF engine to use, the "calibre" engine is recommended as it has automatic header and footer removal. Choices: {}').format(
                ', '.join(ENGINES)
            ),
        ),
        OptionRecommendation(
            name='pdf_header_skip',
            recommended_value=-1,
            help=_(
                'Skip everything to the specified number of pixels at the top of a page.'
                ' Negative numbers mean auto-detect and remove headers, zero means do not remove headers and positive numbers'
                ' mean remove headers that appear above that many pixels from the top of the page. Works only'
                ' with the new PDF engine.'
            ),
        ),
        OptionRecommendation(
            name='pdf_footer_skip',
            recommended_value=-1,
            help=_(
                'Skip everything to the specified number of pixels at the bottom of a page.'
                ' Negative numbers mean auto-detect and remove footers, zero means do not remove footers and positive numbers'
                ' mean remove footers that appear below that many pixels from the bottom of the page. Works only'
                ' with the calibre PDF engine.'
            ),
        ),
        OptionRecommendation(
            name='pdf_header_regex',
            recommended_value='',
            help=_(
                'Regular expression to remove lines at the top of a page. '
                'This only looks at the first line of a page and works only with the calibre PDF engine.'
            ),
        ),
        OptionRecommendation(
            name='pdf_footer_regex',
            recommended_value='',
            help=_(
                'Regular expression to remove lines at the bottom of a page. '
                'This only looks at the last line of a page and works only with the calibre PDF engine.'
            ),
        ),
    }

    def convert(self, stream, options, file_ext, log, accelerators):
        from calibre.ebooks.metadata.opf2 import OPFCreator
        from calibre.ebooks.pdf.pdftohtml import pdftohtml

        log.debug('Converting file to html...')
        # The main html file will be named index.html
        self.opts, self.log = options, log
        if options.pdf_engine == 'calibre':
            from calibre.ebooks.pdf.reflow import PDFDocument
            from calibre.utils.cleantext import clean_ascii_chars

            pdftohtml(os.getcwd(), stream.name, self.opts.no_images, as_xml=True)
            with open('index.xml', 'rb') as f:
                xml = clean_ascii_chars(f.read())
            PDFDocument(xml, self.opts, self.log)
        else:
            pdftohtml(os.getcwd(), stream.name, options.no_images)

        from calibre.ebooks.metadata.meta import get_metadata

        log.debug('Retrieving document metadata...')
        mi = get_metadata(stream, 'pdf')
        opf = OPFCreator(os.getcwd(), mi)

        manifest = [('index.html', None)]

        images = os.listdir(os.getcwd())
        images.remove('index.html')
        for i in images:
            manifest.append((i, None))
        log.debug('Generating manifest...')
        opf.create_manifest(manifest)

        opf.create_spine(['index.html'])
        log.debug('Rendering manifest...')
        with open('metadata.opf', 'wb') as opffile:
            opf.render(opffile)
        if os.path.exists('toc.ncx'):
            ncxid = opf.manifest.id_for_path('toc.ncx')
            if ncxid:
                with open('metadata.opf', 'r+b') as f:
                    raw = f.read().replace(b'<spine', b'<spine toc="%s"' % as_bytes(ncxid))
                    f.seek(0)
                    f.write(raw)

        return os.path.join(os.getcwd(), 'metadata.opf')



# Form implementation generated from reading ui file '/home/kovid/work/calibre/src/calibre/gui2/preferences/look_feel_tabs/bookshelf_view.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_bookshelf_tab(object):
    def setupUi(self, bookshelf_tab):
        bookshelf_tab.setObjectName("bookshelf_tab")
        bookshelf_tab.resize(626, 521)
        self.bookshelf_view_tab = QtWidgets.QWidget()
        self.bookshelf_view_tab.setObjectName("bookshelf_view_tab")
        self.formLayout_3 = QtWidgets.QFormLayout(self.bookshelf_view_tab)
        self.formLayout_3.setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout_3.setObjectName("formLayout_3")
        self.label = QtWidgets.QLabel(parent=self.bookshelf_view_tab)
        self.label.setObjectName("label")
        self.formLayout_3.setWidget(2, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label)
        self.opt_bookshelf_height = QtWidgets.QSpinBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_height.setMinimum(119)
        self.opt_bookshelf_height.setMaximum(1200)
        self.opt_bookshelf_height.setObjectName("opt_bookshelf_height")
        self.formLayout_3.setWidget(2, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_bookshelf_height)
        self.label_variable_height = QtWidgets.QLabel(parent=self.bookshelf_view_tab)
        self.label_variable_height.setObjectName("label_variable_height")
        self.formLayout_3.setWidget(4, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_variable_height)
        self.opt_bookshelf_variable_height = QtWidgets.QComboBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_variable_height.setObjectName("opt_bookshelf_variable_height")
        self.formLayout_3.setWidget(4, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_bookshelf_variable_height)
        self.label_fade_time = QtWidgets.QLabel(parent=self.bookshelf_view_tab)
        self.label_fade_time.setObjectName("label_fade_time")
        self.formLayout_3.setWidget(3, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_fade_time)
        self.opt_bookshelf_fade_time = QtWidgets.QSpinBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_fade_time.setMaximum(2000)
        self.opt_bookshelf_fade_time.setSingleStep(100)
        self.opt_bookshelf_fade_time.setProperty("value", 200)
        self.opt_bookshelf_fade_time.setObjectName("opt_bookshelf_fade_time")
        self.formLayout_3.setWidget(3, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_bookshelf_fade_time)
        self.label_hover_interact = QtWidgets.QLabel(parent=self.bookshelf_view_tab)
        self.label_hover_interact.setObjectName("label_hover_interact")
        self.formLayout_3.setWidget(5, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_hover_interact)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.opt_bookshelf_hover = QtWidgets.QComboBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_hover.setObjectName("opt_bookshelf_hover")
        self.horizontalLayout_6.addWidget(self.opt_bookshelf_hover)
        self.opt_bookshelf_shadow = QtWidgets.QCheckBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_shadow.setObjectName("opt_bookshelf_shadow")
        self.horizontalLayout_6.addWidget(self.opt_bookshelf_shadow)
        self.formLayout_3.setLayout(5, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_6)
        self.label_4 = QtWidgets.QLabel(parent=self.bookshelf_view_tab)
        self.label_4.setObjectName("label_4")
        self.formLayout_3.setWidget(7, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_4)
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.opt_bookshelf_divider_style = QtWidgets.QComboBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_divider_style.setObjectName("opt_bookshelf_divider_style")
        self.horizontalLayout_8.addWidget(self.opt_bookshelf_divider_style)
        self.opt_bookshelf_divider_text_right = QtWidgets.QCheckBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_divider_text_right.setObjectName("opt_bookshelf_divider_text_right")
        self.horizontalLayout_8.addWidget(self.opt_bookshelf_divider_text_right)
        self.opt_bookshelf_start_with_divider = QtWidgets.QCheckBox(parent=self.bookshelf_view_tab)
        self.opt_bookshelf_start_with_divider.setObjectName("opt_bookshelf_start_with_divider")
        self.horizontalLayout_8.addWidget(self.opt_bookshelf_start_with_divider)
        self.formLayout_3.setLayout(7, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_8)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.recount_button = QtWidgets.QPushButton(parent=self.bookshelf_view_tab)
        self.recount_button.setObjectName("recount_button")
        self.horizontalLayout_7.addWidget(self.recount_button)
        self.show_log_button = QtWidgets.QPushButton(parent=self.bookshelf_view_tab)
        self.show_log_button.setObjectName("show_log_button")
        self.horizontalLayout_7.addWidget(self.show_log_button)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_7.addItem(spacerItem)
        self.formLayout_3.setLayout(9, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.horizontalLayout_7)
        spacerItem1 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.formLayout_3.setItem(10, QtWidgets.QFormLayout.ItemRole.LabelRole, spacerItem1)
        self.groupBox = QtWidgets.QGroupBox(parent=self.bookshelf_view_tab)
        self.groupBox.setObjectName("groupBox")
        self.formLayout = QtWidgets.QFormLayout(self.groupBox)
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.FieldGrowthPolicy.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label_bookshelf_thumbnail = QtWidgets.QLabel(parent=self.groupBox)
        self.label_bookshelf_thumbnail.setObjectName("label_bookshelf_thumbnail")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_bookshelf_thumbnail)
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.opt_bookshelf_thumbnail = QtWidgets.QComboBox(parent=self.groupBox)
        self.opt_bookshelf_thumbnail.setToolTip("")
        self.opt_bookshelf_thumbnail.setObjectName("opt_bookshelf_thumbnail")
        self.horizontalLayout_10.addWidget(self.opt_bookshelf_thumbnail)
        self.label_6 = QtWidgets.QLabel(parent=self.groupBox)
        self.label_6.setObjectName("label_6")
        self.horizontalLayout_10.addWidget(self.label_6)
        self.opt_bookshelf_thumbnail_opacity = QtWidgets.QSpinBox(parent=self.groupBox)
        self.opt_bookshelf_thumbnail_opacity.setMinimum(0)
        self.opt_bookshelf_thumbnail_opacity.setMaximum(100)
        self.opt_bookshelf_thumbnail_opacity.setProperty("value", 30)
        self.opt_bookshelf_thumbnail_opacity.setObjectName("opt_bookshelf_thumbnail_opacity")
        self.horizontalLayout_10.addWidget(self.opt_bookshelf_thumbnail_opacity)
        self.formLayout.setLayout(0, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_10)
        self.label_title = QtWidgets.QLabel(parent=self.groupBox)
        self.label_title.setObjectName("label_title")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_title)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.opt_bookshelf_title_template = QtWidgets.QLineEdit(parent=self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(10)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.opt_bookshelf_title_template.sizePolicy().hasHeightForWidth())
        self.opt_bookshelf_title_template.setSizePolicy(sizePolicy)
        self.opt_bookshelf_title_template.setClearButtonEnabled(True)
        self.opt_bookshelf_title_template.setObjectName("opt_bookshelf_title_template")
        self.horizontalLayout_2.addWidget(self.opt_bookshelf_title_template)
        self.template_title_button = QtWidgets.QPushButton(parent=self.groupBox)
        self.template_title_button.setObjectName("template_title_button")
        self.horizontalLayout_2.addWidget(self.template_title_button)
        self.formLayout.setLayout(1, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_2)
        self.label_2 = QtWidgets.QLabel(parent=self.groupBox)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_2)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.opt_bookshelf_author_template = QtWidgets.QLineEdit(parent=self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(10)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.opt_bookshelf_author_template.sizePolicy().hasHeightForWidth())
        self.opt_bookshelf_author_template.setSizePolicy(sizePolicy)
        self.opt_bookshelf_author_template.setClearButtonEnabled(True)
        self.opt_bookshelf_author_template.setObjectName("opt_bookshelf_author_template")
        self.horizontalLayout_4.addWidget(self.opt_bookshelf_author_template)
        self.template_author_button = QtWidgets.QPushButton(parent=self.groupBox)
        self.template_author_button.setObjectName("template_author_button")
        self.horizontalLayout_4.addWidget(self.template_author_button)
        self.formLayout.setLayout(2, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_4)
        self.label_pages = QtWidgets.QLabel(parent=self.groupBox)
        self.label_pages.setObjectName("label_pages")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_pages)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.opt_bookshelf_spine_size_template = QtWidgets.QLineEdit(parent=self.groupBox)
        self.opt_bookshelf_spine_size_template.setToolTip("")
        self.opt_bookshelf_spine_size_template.setClearButtonEnabled(True)
        self.opt_bookshelf_spine_size_template.setObjectName("opt_bookshelf_spine_size_template")
        self.horizontalLayout_3.addWidget(self.opt_bookshelf_spine_size_template)
        self.template_pages_button = QtWidgets.QPushButton(parent=self.groupBox)
        self.template_pages_button.setObjectName("template_pages_button")
        self.horizontalLayout_3.addWidget(self.template_pages_button)
        self.use_pages_button = QtWidgets.QPushButton(parent=self.groupBox)
        self.use_pages_button.setObjectName("use_pages_button")
        self.horizontalLayout_3.addWidget(self.use_pages_button)
        self.formLayout.setLayout(3, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_3)
        self.label_10 = QtWidgets.QLabel(parent=self.groupBox)
        self.label_10.setObjectName("label_10")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_10)
        self.label_3 = QtWidgets.QLabel(parent=self.groupBox)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_3)
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.label_7 = QtWidgets.QLabel(parent=self.groupBox)
        self.label_7.setObjectName("label_7")
        self.horizontalLayout_9.addWidget(self.label_7)
        self.opt_bookshelf_min_font_multiplier = QtWidgets.QDoubleSpinBox(parent=self.groupBox)
        self.opt_bookshelf_min_font_multiplier.setDecimals(2)
        self.opt_bookshelf_min_font_multiplier.setMinimum(0.1)
        self.opt_bookshelf_min_font_multiplier.setMaximum(1.0)
        self.opt_bookshelf_min_font_multiplier.setSingleStep(0.1)
        self.opt_bookshelf_min_font_multiplier.setObjectName("opt_bookshelf_min_font_multiplier")
        self.horizontalLayout_9.addWidget(self.opt_bookshelf_min_font_multiplier)
        self.label_8 = QtWidgets.QLabel(parent=self.groupBox)
        self.label_8.setObjectName("label_8")
        self.horizontalLayout_9.addWidget(self.label_8)
        self.opt_bookshelf_max_font_multiplier = QtWidgets.QDoubleSpinBox(parent=self.groupBox)
        self.opt_bookshelf_max_font_multiplier.setMinimum(1.0)
        self.opt_bookshelf_max_font_multiplier.setMaximum(3.0)
        self.opt_bookshelf_max_font_multiplier.setSingleStep(0.1)
        self.opt_bookshelf_max_font_multiplier.setObjectName("opt_bookshelf_max_font_multiplier")
        self.horizontalLayout_9.addWidget(self.opt_bookshelf_max_font_multiplier)
        self.label_9 = QtWidgets.QLabel(parent=self.groupBox)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_9.addWidget(self.label_9)
        self.opt_bookshelf_outline_width = QtWidgets.QDoubleSpinBox(parent=self.groupBox)
        self.opt_bookshelf_outline_width.setDecimals(1)
        self.opt_bookshelf_outline_width.setMaximum(3.0)
        self.opt_bookshelf_outline_width.setSingleStep(0.1)
        self.opt_bookshelf_outline_width.setObjectName("opt_bookshelf_outline_width")
        self.horizontalLayout_9.addWidget(self.opt_bookshelf_outline_width)
        self.formLayout.setLayout(5, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_9)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.opt_bookshelf_up_to_down = QtWidgets.QCheckBox(parent=self.groupBox)
        self.opt_bookshelf_up_to_down.setObjectName("opt_bookshelf_up_to_down")
        self.horizontalLayout.addWidget(self.opt_bookshelf_up_to_down)
        self.opt_bookshelf_make_space_for_second_line = QtWidgets.QCheckBox(parent=self.groupBox)
        self.opt_bookshelf_make_space_for_second_line.setObjectName("opt_bookshelf_make_space_for_second_line")
        self.horizontalLayout.addWidget(self.opt_bookshelf_make_space_for_second_line)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.formLayout.setLayout(7, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.horizontalLayout)
        self.horizontalLayout_14 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_14.setObjectName("horizontalLayout_14")
        self.bookshelf_font_display = QtWidgets.QLineEdit(parent=self.groupBox)
        self.bookshelf_font_display.setReadOnly(True)
        self.bookshelf_font_display.setObjectName("bookshelf_font_display")
        self.horizontalLayout_14.addWidget(self.bookshelf_font_display)
        self.change_font_button = QtWidgets.QPushButton(parent=self.groupBox)
        self.change_font_button.setObjectName("change_font_button")
        self.horizontalLayout_14.addWidget(self.change_font_button)
        self.restore_default_font_button = QtWidgets.QPushButton(parent=self.groupBox)
        icon = QtGui.QIcon.ic("clear_left.png")
        self.restore_default_font_button.setIcon(icon)
        self.restore_default_font_button.setObjectName("restore_default_font_button")
        self.horizontalLayout_14.addWidget(self.restore_default_font_button)
        self.formLayout.setLayout(4, QtWidgets.QFormLayout.ItemRole.FieldRole, self.horizontalLayout_14)
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.groupBox)
        bookshelf_tab.addTab(self.bookshelf_view_tab, "")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.formLayout_2 = QtWidgets.QFormLayout(self.tab)
        self.formLayout_2.setObjectName("formLayout_2")
        self.label_5 = QtWidgets.QLabel(parent=self.tab)
        self.label_5.setObjectName("label_5")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.ItemRole.LabelRole, self.label_5)
        self.opt_bookshelf_theme_override = QtWidgets.QComboBox(parent=self.tab)
        self.opt_bookshelf_theme_override.setObjectName("opt_bookshelf_theme_override")
        self.formLayout_2.setWidget(0, QtWidgets.QFormLayout.ItemRole.FieldRole, self.opt_bookshelf_theme_override)
        self.opt_bookshelf_use_custom_background = QtWidgets.QCheckBox(parent=self.tab)
        self.opt_bookshelf_use_custom_background.setObjectName("opt_bookshelf_use_custom_background")
        self.formLayout_2.setWidget(1, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_bookshelf_use_custom_background)
        self.background_box = BackgroundConfig(parent=self.tab)
        self.background_box.setObjectName("background_box")
        self.formLayout_2.setWidget(2, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.background_box)
        self.opt_bookshelf_use_custom_colors = QtWidgets.QGroupBox(parent=self.tab)
        self.opt_bookshelf_use_custom_colors.setCheckable(True)
        self.opt_bookshelf_use_custom_colors.setChecked(False)
        self.opt_bookshelf_use_custom_colors.setObjectName("opt_bookshelf_use_custom_colors")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.opt_bookshelf_use_custom_colors)
        self.verticalLayout.setObjectName("verticalLayout")
        self.bookshelf_tab = QtWidgets.QTabWidget(parent=self.opt_bookshelf_use_custom_colors)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.bookshelf_tab.sizePolicy().hasHeightForWidth())
        self.bookshelf_tab.setSizePolicy(sizePolicy)
        self.bookshelf_tab.setObjectName("bookshelf_tab")
        self.widget_1 = QtWidgets.QWidget()
        self.widget_1.setObjectName("widget_1")
        self.custom_colors_light_layout = QtWidgets.QFormLayout(self.widget_1)
        self.custom_colors_light_layout.setObjectName("custom_colors_light_layout")
        self.bookshelf_tab.addTab(self.widget_1, "")
        self.widget_2 = QtWidgets.QWidget()
        self.widget_2.setObjectName("widget_2")
        self.custom_colors_dark_layout = QtWidgets.QFormLayout(self.widget_2)
        self.custom_colors_dark_layout.setObjectName("custom_colors_dark_layout")
        self.bookshelf_tab.addTab(self.widget_2, "")
        self.verticalLayout.addWidget(self.bookshelf_tab)
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.restore_defaults_colors_button = QtWidgets.QPushButton(parent=self.opt_bookshelf_use_custom_colors)
        self.restore_defaults_colors_button.setIcon(icon)
        self.restore_defaults_colors_button.setObjectName("restore_defaults_colors_button")
        self.horizontalLayout_11.addWidget(self.restore_defaults_colors_button)
        spacerItem3 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.horizontalLayout_11.addItem(spacerItem3)
        self.verticalLayout.addLayout(self.horizontalLayout_11)
        self.formLayout_2.setWidget(3, QtWidgets.QFormLayout.ItemRole.SpanningRole, self.opt_bookshelf_use_custom_colors)
        bookshelf_tab.addTab(self.tab, "")
        self.opt_bookshelf_icon_rules = EditRulesWidget()
        self.opt_bookshelf_icon_rules.setObjectName("opt_bookshelf_icon_rules")
        bookshelf_tab.addTab(self.opt_bookshelf_icon_rules, "")
        self.config_cache = CoverCacheConfig()
        self.config_cache.setObjectName("config_cache")
        bookshelf_tab.addTab(self.config_cache, "")
        self.label.setBuddy(self.opt_bookshelf_height)
        self.label_variable_height.setBuddy(self.opt_bookshelf_variable_height)
        self.label_fade_time.setBuddy(self.opt_bookshelf_fade_time)
        self.label_hover_interact.setBuddy(self.opt_bookshelf_hover)
        self.label_4.setBuddy(self.opt_bookshelf_divider_style)
        self.label_bookshelf_thumbnail.setBuddy(self.opt_bookshelf_thumbnail)
        self.label_6.setBuddy(self.opt_bookshelf_thumbnail_opacity)
        self.label_title.setBuddy(self.opt_bookshelf_title_template)
        self.label_2.setBuddy(self.opt_bookshelf_author_template)
        self.label_pages.setBuddy(self.opt_bookshelf_spine_size_template)
        self.label_10.setBuddy(self.bookshelf_font_display)
        self.label_3.setBuddy(self.opt_bookshelf_min_font_multiplier)
        self.label_7.setBuddy(self.opt_bookshelf_min_font_multiplier)
        self.label_8.setBuddy(self.opt_bookshelf_max_font_multiplier)
        self.label_9.setBuddy(self.opt_bookshelf_outline_width)
        self.label_5.setBuddy(self.opt_bookshelf_theme_override)

        self.retranslateUi(bookshelf_tab)
        bookshelf_tab.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(bookshelf_tab)

    def retranslateUi(self, bookshelf_tab):

        self.label.setText(_("Shelf &height:"))
        self.opt_bookshelf_height.setToolTip(_("Control the height of the book shelves, in pixels. The default is to choose a height automatically based on the size of the screen. Note that if you change the height, the cover thumbnail cache will need to be re-populated so scrolling might be a little slow to start with."))
        self.opt_bookshelf_height.setSpecialValueText(_("Automatic"))
        self.opt_bookshelf_height.setSuffix(_(" px"))
        self.label_variable_height.setText(_("Book &height variation:"))
        self.opt_bookshelf_variable_height.setToolTip(_("Control how the height of book spines varies on the shelf. \"Constant\" makes all books the same height. \"Constant per group\" makes all books in the same group the same height. \"Variable\" gives each book a unique height."))
        self.label_fade_time.setText(_("Duration of hover &animation:"))
        self.opt_bookshelf_fade_time.setSpecialValueText(_("Disable"))
        self.opt_bookshelf_fade_time.setSuffix(_(" milliseconds"))
        self.label_hover_interact.setText(_("&Expand the hovered cover:"))
        self.opt_bookshelf_hover.setToolTip(_("Control what happens when the mouse hovers over the spine of a book."))
        self.opt_bookshelf_shadow.setText(_("Show &shadow for hovered cover"))
        self.label_4.setText(_("Group divider &style:"))
        self.opt_bookshelf_divider_style.setToolTip(_("Control the style to use for the group divider"))
        self.opt_bookshelf_divider_text_right.setToolTip(_("Control whether the text of the group divider is aligned to the right or left"))
        self.opt_bookshelf_divider_text_right.setText(_("Text &aligned to the right"))
        self.opt_bookshelf_start_with_divider.setToolTip(_("When grouping books and a group extends to more than one shelf, place a group divider at the beginning of the additional shelves"))
        self.opt_bookshelf_start_with_divider.setText(_("&Start shelves with divider"))
        self.recount_button.setToolTip(_("Have calibre re-count the pages in all books in the current library."))
        self.recount_button.setText(_("&Recount pages"))
        self.show_log_button.setText(_("Show count &failure log"))
        self.groupBox.setTitle(_("Book spine"))
        self.label_bookshelf_thumbnail.setText(_("Show &cover:"))
        self.label_6.setText(_("&Opacity:"))
        self.opt_bookshelf_thumbnail_opacity.setToolTip(_("<p>How much to blend the cover into the spine. Higher values decrease blending."))
        self.opt_bookshelf_thumbnail_opacity.setSuffix(_(" %"))
        self.label_title.setText(_("Template for &first line:"))
        self.opt_bookshelf_title_template.setToolTip(_("<p>The template used to generate the main text on the spine. This template uses the same syntax as save templates. Defaults to just the book title. Note that this setting is per-library, which means that you have to set it again for every different calibre library you use.\n"
"<p>For example, if you want the first line to be the book title in upper case, you can use the template: <code>{title:uppercase()}</code>"))
        self.template_title_button.setText(_("Template &editor"))
        self.label_2.setText(_("Template for &second line:"))
        self.opt_bookshelf_author_template.setToolTip(_("<p>The template used to generate an optional second line of text on the spine. This template uses the same syntax as save templates. Defaults to no second line. For example, use the value {authors} to show the book authors. Note that this setting is per-library, which means that you have to set it again for every different calibre library you use.</p>"))
        self.template_author_button.setText(_("Template &editor"))
        self.label_pages.setText(_("Template for &width:"))
        self.template_pages_button.setText(_("Template &editor"))
        self.use_pages_button.setText(_("Use a custom &pages column"))
        self.label_10.setText(_("&Font:"))
        self.label_3.setText(_("&Font size:"))
        self.label_7.setText(_("&Minimum:"))
        self.opt_bookshelf_min_font_multiplier.setToolTip(_("The minimum font size for text on the spine. It is a fraction of the base font size."))
        self.label_8.setText(_("Ma&ximum:"))
        self.opt_bookshelf_max_font_multiplier.setToolTip(_("The maximum font size for text on the spine. It is a multiple of the base font size."))
        self.label_9.setText(_("&Outline"))
        self.opt_bookshelf_outline_width.setToolTip(_("<p>Set the thickness, in pixels, of the outline (border) around text on the spine. A value of zero means no outline. An outline helps the text standout against the spine, especially when the spine has a colorful cover."))
        self.opt_bookshelf_outline_width.setSpecialValueText(_("none"))
        self.opt_bookshelf_outline_width.setSuffix(_(" px"))
        self.opt_bookshelf_up_to_down.setToolTip(_("Control whether the text shown on book spines goes upwards or downwards"))
        self.opt_bookshelf_up_to_down.setText(_("Spine text &downwards"))
        self.opt_bookshelf_make_space_for_second_line.setToolTip(_("<p>When showing a second line in the spine, for very narrow books, artificially widen the book to make space for the second line. Without this, the second line will not be displayed for narrow books, with it, narrow books will appear artificially wider than they really are."))
        self.opt_bookshelf_make_space_for_second_line.setText(_("Make &space for second line"))
        self.bookshelf_font_display.setPlaceholderText(_("Use default font"))
        self.change_font_button.setText(_("&Change"))
        self.restore_default_font_button.setText(_("Restore &default"))
        bookshelf_tab.setTabText(bookshelf_tab.indexOf(self.bookshelf_view_tab), _("&Layout"))
        self.label_5.setText(_("&Dark/light:"))
        self.opt_bookshelf_theme_override.setToolTip(_("Control the color theme to use for the bookshelf view"))
        self.opt_bookshelf_use_custom_background.setToolTip(_("Use a custom background, either color or image"))
        self.opt_bookshelf_use_custom_background.setText(_("Use a custom &background"))
        self.opt_bookshelf_use_custom_colors.setTitle(_("Use customs &colors"))
        self.bookshelf_tab.setTabText(self.bookshelf_tab.indexOf(self.widget_1), _("&Light mode"))
        self.bookshelf_tab.setTabText(self.bookshelf_tab.indexOf(self.widget_2), _("&Dark mode"))
        self.restore_defaults_colors_button.setText(_("Restore &defaults"))
        bookshelf_tab.setTabText(bookshelf_tab.indexOf(self.tab), _("&Colors"))
        bookshelf_tab.setTabText(bookshelf_tab.indexOf(self.opt_bookshelf_icon_rules), _("&Emblems"))
        bookshelf_tab.setTabText(bookshelf_tab.indexOf(self.config_cache), _("&Performance"))
from calibre.gui2.preferences.look_feel_tabs import BackgroundConfig, CoverCacheConfig, EditRulesWidget
